import numpy as np
import numpy.linalg
import itertools
import os
from collections import defaultdict
import networkx as nx
from scipy.spatial import ConvexHull
from pulp import *

from qmpy.utils import *
from phase import Phase, PhaseData
from reaction import Reaction

class PhaseSpaceError(Exception):
    pass

class PhaseSpace(object):
    '''
    A PhaseSpace object represents, naturally, a region of phase space. 
    
    The most fundamental property of a PhaseSpace is its bounds,
    which are given as a hyphen-delimited list of compositions. These represent
    the extent of the phase space, and determine which phases are within the
    space.

    Next, a PhaseSpace has an attribute, _data, which is a PhaseData object,
    and is a container for Phase objects, which are used when performing
    thermodynamic analysis on this space.

    The majority of attributes are lazy, that is, they are only computed when
    they are requested, and how to get them (of which there are often several
    ways) is decided based on the size and shape of the phase space.
    '''
    
    def __init__(self, definition=None, pdata=PhaseData(), load=False):
        self._data = pdata
        self._phases = None # phases within the space
        self._phase_dict = None # phase dict
        self._space = None # set of elements present
        self._hull = None # list of facets on hull; facets are lists of phases
        self._stable = None # list of phases that are stable
        self._tie_lines = None # list of pairs connected by tie lines
        self._spaces = None # the biggest spaces containing phases
        self._dual_spaces = None # the biggest spaces containing tie lines
        self._cliques = None # maximal k-cliques in space 
        self._graph = None # networkx graph of phase space

        if definition is None:
            self.bounds = None
            return

        if isinstance(definition, basestring):
            if '-' in definition:
                definition = [ parse_comp(b) for b in definition.split('-')]
            elif ' ' in definition:
                definition = [ parse_comp(b) for b in definition.split(' ')]
        self.bounds = definition

        if load is None:
            pass
        elif not pdata._phases and not load:
            self.load('oqmd')
        elif load:
            self.load(*load)

        self.get_phases()
        self.get_phase_dict()

    ####
    #### Builtins
    ####

    def __repr__(self):
        if self._bounds is None:
            return '<unbounded PhaseSpace>'
        names = [ format_comp(reduce_comp(b)) for b in self._raw_bounds ]
        return '<PhaseSpace bound by %s>' % '-'.join(names)

    ###
    ### Data access and management
    ###

    def clear_data(self):
        '''
        Clears all phase data:
            PhaseSpace._data = PhaseData()
            PhasesSpace._phases = None
            PhaseSpace._phase_dict = None
        '''
        self._data._phases = []
        self._data._phase_dict = None
        self._phases = None
        self._phase_dict = None

    def clear_analysis(self):
        '''
        Clears all analyzed results.
            PhaseSpace._space = None
            PhaseSpace._stable = None
            PhaseSpace._tie_lines = None
            PhaseSpace._hull = None
            PhaseSpace._graph = None
        '''
        self._stable = None
        self._tie_lines = None
        self._hull = None
        self._spaces = None
        self._dual_spaces = None
        self._cliques = None
        self._graph = None

    def clear_all(self):
        '''
        Clears input data and analyzed results. 
        Same as:
        >>> PhaseData.clear_data() 
        >>> PhaseData.clear_analysis()
        '''
        self.clear_data()
        self.clear_analysis()

    def load(self, *args, **kwargs):
        '''
        Load OQMD data into PhaseData. Takes unordered arguments; 
            if 'legacy' in args: will import older data
            if 'stable' in args: will only import stable structures
            if 'prototypes' in args: will import prototype structures
            if 'icsd' in args: will import icsd structures
            if 'oqmd' in args: will import all oqmd structures
            if 'get=[string]' in args: will import structures with [string] in
            their path.
        '''
        stable = ( 'stable' in args )
        fit = kwargs.get('fit', 'standard')
        for arg in args:
            if arg == 'legacy':
                self._data.load('legacy.dat')
            if arg == 'oqmd':
                self._data.load_oqmd(self.space)
            if arg == 'icsd':
                self._data.load_oqmd(self.space,
                        search={'path__contains':'icsd'},
                        stable=stable)
            if arg == 'prototypes':
                self._data.load_oqmd(space=self.space,
                        search={'path__contains':'prototypes'},
                        stable=stable)
            if arg.startswith('get='):
                get = arg.replace('get=','')
                self._data.load_oqmd(space=self.space,
                        search={'path__contains':get},
                        stable=stable,
                        fit=fit)

    def load_tie_lines(self):
        raise NotImplementedError

    ####
    #### Properties of the space itself
    ####

    @property
    def bounds(self):
        return self._bounds

    @bounds.setter
    def bounds(self, bound_list):
        if bound_list is None:
            self._bounds = None
            self._orth = None
            self._raw_bounds = None
            return
        self._raw_bounds = bound_list
        self._bounds = []
        basis = []
        for b in bound_list:
            b = defaultdict(float, b)
            b_tot = float(sum(b.values()))
            self._bounds.append(
                    dict((k, v/b_tot) for k, v in b.items()))
            basis.append([ b[k]/b_tot for k in self.sorted_space])
        self._basis = np.array(basis)

    @property
    def dimension(self):
        '''
        Number of bounds; note this is distinct from the number of elements.

        Fe-Li-O is 2-D 
        FeO-Ni2O-CoO-Ti3O4 is 4-D
        FeO-Fe2O3-Fe is 3-D
        '''
        return len(self.bounds)

    @property
    def space(self):
        '''
        Set of elements present in the PhaseSpace.

        Fe-Si-Li -> set(['Fe', 'Li', 'Si'])
        FeO-NiO-Ti2O3 -> set(['Ti', 'Ni', 'O', 'Fe'])
        '''

        if self._space is None:
            self.get_space()
        return self._space

    @property
    def sorted_space(self):
        return sorted(self.space)

    ###
    ### Properties of the phases within the space
    ###

    @property
    def phases(self):
        '''
        list of phases in space
        '''
        if not self._phases:
            self.get_phases()
        return self._phases

    @property
    def phase_dict(self):
        '''
        dictionary of phases (keys are phase names)
        '''
        if self._phase_dict is None:
            self.get_phase_dict()
        return self._phase_dict

    @property
    def spaces(self):
        '''
        list of sets of elements, such that every phase in self.phases
        is contained in at least one set, and no set is a subset of
        any other.
        '''
        if self._spaces is None:
            self.get_spaces()
        return self._spaces

    @property
    def stable(self):
        '''
        list of stable phases
        '''
        if self._stable is None:
            self.get_space_hull()
        return self._stable

    @property
    def unstable(self):
        '''
        list of unstable phases.
        '''
        if self._stable is None:
            self.get_space_hull()
        return [ p for p in self.phase_dict.values() if
            not p in self._stable and self.in_space(p) ]

    @property
    def tie_lines(self):
        '''
        list of length 2 tuples of phases with tie lines between them
        '''
        if self._tie_lines is None:
            self.get_tie_lines()
        return self._tie_lines

    @property
    def hull(self):
        '''
        List of phases on convex hull within space
        '''
        if self._hull is None:
            if any( len(b) > 1 for b in self.bounds ):
                points = self.get_hull_points()
                self.get_qhull(phases=points)
            else:
                self.get_qhull()
        return self._hull

    @property
    def dual_spaces(self):
        '''
        list of sets of elements, such that any possible tie-line
        between two phases in phases is contained in at least one
        set, and no set is a subset of any other.
        '''
        if self._dual_spaces is None:
            self.get_dual_spaces()
        return self._dual_spaces

    @property
    def graph(self):
        '''
        networkx.Graph representation of the phase space.
        '''
        if self._graph is None:
            self.get_graph()
        return self._graph

    @property
    def cliques(self):
        '''
        Iterator over maximal cliques in the phase space. To get a list of
        cliques, use list(PhaseSpace.cliques).
        '''
        if self._cliques is None:
            self.get_maximal_cliques()
        return self._cliques

    ###
    ### property getters
    ###

    def get_space(self):
        space = set()
        if self.bounds is None:
            return None
        for b in self._raw_bounds:
            space |= set(b.keys())
        self._space = space

    def get_phases(self):
        self._phases = [ p for p in self._data.phases 
                if self.in_space(p) ]

    def get_phase_dict(self):
        phase_dict = {}
        for p in self._data.phase_dict.values():
            if self.in_space(p):
                phase_dict[p.name] = p
        self._phase_dict = phase_dict

    def get_spaces(self):
        spaces = [ p.space for p in self.phase_dict.values() ]
        spaces = set([ frozenset(space) for space in spaces if not 
                any( space < space2 for space2 in spaces ) ])
        self._spaces = spaces

    def get_tie_lines(self):
        if any( len(b) > 1 for b in self.bounds ):
            points = self.get_hull_points()
            self.get_qhull(phases=points)
        else:
            self.get_qhull()
            #self.get_dual_space_hull()

    def get_dual_spaces(self):
        if len(self.spaces) == 1:
            self._dual_spaces = self._spaces
            return

        dual_spaces = []
        imax = len(self.spaces)**2 / 2
        for i, c1 in enumerate(itertools.combinations(self.spaces, r=2)):
            for j, c2 in enumerate(dual_spaces):
                if c1 <= c2:
                    break
                elif c2 < c1:
                    dual_spaces[j] = c1
                    break
            else:
                dual_spaces.append(c1)

        second_pass = []
        imax = len(dual_spaces)
        for i, c1 in enumerate(dual_spaces):
            for c2 in list(dual_spaces):
                if c1 <= c2:
                    break
            else:
                second_pass.append(c1)

        dual_spaces = set(second_pass)
        #dual_spaces = set( frozenset(c1 | c2) for c1, c2 in
        #        itertools.combinations(spaces, r=2) )

        dual_spaces = set([ space for space in dual_spaces 
            if not any( space < space2 for space2 in dual_spaces) ])

        dual_spaces = set([ frozenset([ ids[elt] for elt in space ]) for space
            in dual_spaces])

        self._dual_spaces = dual_spaces

    def get_maximal_cliques(self):
        graph = nx.Graph()
        graph.add_edges_from(self.tie_lines)
        self._cliques = list(nx.find_cliques(graph))
    
    def get_graph(self):
        graph = nx.Graph()
        for n in self.stable:
            graph.add_node('$\\rm{%s}$' % n.latex)
        for t in self.tie_lines:
            t = list(t)
            graph.add_edge('$\\rm{%s}$' % t[0].latex, 
                    '$\\rm{%s}$' % t[1].latex)
        self._graph = graph

    ### hull finding methods

    def get_space_hull(self):
        '''
        Iterate over the maximal spaces, getting the hull of each space. 

        Sets the following variables:
            - stable (complete)

        Advantages:
            - Gets the stable phases quickly

        Disadvantages:
            - Misses a lot of tie lines and facets
        '''
        stable = set()
        tie_lines = set()
        hull = set()
        imax = len(self.spaces)-1
        i = 0
        for space in self.spaces:
            print 'space hull %s/%s : %s' % (i, imax, space)
            i += 1
            space = PhaseSpace('-'.join(space), pdata=self._data, load=None)
            space.get_qhull()

            stable |= space.stable
            tie_lines |= space.tie_lines
            hull |= space.hull
        self._hull = hull # [ list(face) for face in hull ]
        self._tie_lines = tie_lines
        self._stable = stable

    def get_dual_space_hull(self):
        '''
        Iterate over the maximal dual spaces, getting the hull of each space.

        Sets the following variables:
            - stable (complete)
            - tie_lines (complete)

        Advantages:
            - Very fast list of tie lines

        Disadvantages:
            - Slow(er) for getting stable
        '''
        stable = set()
        tie_lines = set()
        hull = set()
        imax = len(self.dual_spaces)-1
        i = 0
        for space in self.dual_spaces:
            print 'dual space hull %s/%s : %s' % (i, imax, space)
            i += 1
            space = PhaseSpace('-'.join(space), pdata=self._data, load=None)
            space.get_qhull()
            stable |= space.stable
            tie_lines |= space.tie_lines
            hull |= space.hull
        self._stable = list(stable)
        self._tie_lines = [ list(tl) for tl in tie_lines ]
        self_hull = list(hull)

    def get_full_hull(self):
        '''
        Iterate over cliques within the space, getting the hull of each.

        Sets the following variables (requires stable + tie_lines):
            - hull (complete)
        '''
        hull = set()
        tie_lines = set()
        stable = set()
        i = 0
        imax = len(self.cliques)-1
        for clique in self.cliques:
            #print 'full space hull %s/%s' % (i, imax)
            i += 1
            self.get_qhull(phases=clique)

            stable |= self.stable
            tie_lines |= self.tie_lines
            hull |= self.hull
        self._hull = hull # [ list(face) for face in hull ]
        self._tie_lines = tie_lines
        self._stable = stable

    ### Methods using GCLP 

    def gclp(self, composition={}, mus={}, phases=[]):
        '''
        Returns energy, phase composition which is stable at given composition

        Example:
        >>> space = PhaseSpace('Fe-Li-O')
        >>> phases, energy = space.gclp('FeLiO2')
        >>> print phases
        >>> print energy
        '''
        if not composition:
            return 0.0, {}

        if isinstance(composition, basestring):
            composition = parse_comp(composition)
            
        if not phases:
            phases = self.phase_dict.values()

        space = set(composition.keys())
        in_phases = []
        for phase in phases:
            if phase.space <= space:
                in_phases.append(phase)

        return self._gclp(composition=composition,
                mus=mus, phases=in_phases)

    def _gclp(self, composition={}, mus={}, phases=[]):
        prob = LpProblem('GibbsEnergyMin', LpMinimize)
        phase_vars = LpVariable.dicts('lib', phases, 0.0)
        prob += lpSum([ (phase.energy -
            sum([ phase.unit_comp[elt]*mu
                for elt, mu in mus.items() ])) * phase_vars[phase]
            for phase in phases]),\
                    "Free Energy"
        for elt, constraint in composition.items():
            prob += lpSum([ 
                phase.unit_comp.get(elt,0)*phase_vars[phase]
                for phase in phases ]) == float(constraint),\
                        'Conservation of '+elt

        #prob.solve(GUROBI(msg=False))
        #print phases
        #print composition
        prob.solve()

        E = value(prob.objective)
        xsoln = dict([ (phase, phase_vars[phase].varValue)
            for phase in phases if phase_vars[phase].varValue > 1e-5])
        return E, xsoln

    def get_minima(self, phases, bounds):
        '''
        Given a set of Phases, get_minima will determine the minimum
        free energy elemental composition as a weighted sum of these
        compounds
        '''

        prob = LpProblem('GibbsEnergyMin', LpMinimize)
        pvars = LpVariable.dicts('phase', phases, 0)
        bvars = LpVariable.dicts('bound', bounds, 0.0, 1.0)
        prob += lpSum( phase.energy*pvars[phase] for phase in phases ) - \
                lpSum( bound.energy*bvars[bound] for bound in bounds ), \
                                "Free Energy"
        for elt in self.space:
            prob += sum([ p.unit_comp.get(elt,0)*pvars[p] for p in phases ])\
                        == \
                sum([ b.unit_comp.get(elt, 0)*bvars[b] for b in bounds ]),\
                            'Contraint to the proper range of'+elt
        prob += sum([ bvars[b] for b in bounds ]) == 1, \
                'sum of bounds must be 1'

        #prob.solve(GUROBI(msg=False))
        prob.solve()
        E = value(prob.objective)
        xsoln = defaultdict(float,
            [(p, pvars[p].varValue) for p in phases if 
                abs(pvars[p].varValue) > 1e-6])
        return xsoln, E

    def chempot_range(self, phase):
        e = phase.energy
        pot_bounds = {}
        for elt in phase.comp.keys():
            tcomp = dict(phase.unit_comp)
            tcomp[elt] -= 0.01
            eup, xup = self.gclp(tcomp)
            tcomp[elt] += 0.02
            edo, xdo = self.gclp(tcomp)
            pot_bounds[elt] = [ (edo-e)/0.01, (e-eup)/0.01 ]
        return pot_bounds
            

    def get_tie_lines_by_gclp(self, iterable=False):
        tie_lines=[]
        self.get_gclp_stable()

        for k1, k2 in itertools.combinations(self.stable, 2):
            #for k1, k2 in self.tie_lines:
            testpoint = (self.coord(k1.unit_comp) + self.coord(k2.unit_comp))/2
            energy, phases = self.gclp(self.comp(testpoint))
            if abs(energy - (k1.energy + k2.energy)/2) < 1e-8:
                tie_lines.append([k1,k2])
                if iterable:
                    yield [k1, k2]
        self._tie_lines = tie_lines

    def coord(self, composition):
        '''Returns the coordinate of a composition.
        
        Example:
        >>> space = PhaseSpace('Fe-Li-O')
        >>> space.coord({'Fe':1, 'Li':1, 'O':2})
        array([ 0.25,  0.25,  0.5 ])
        '''
        if isinstance(composition, Phase):
            composition = composition.comp
        elif isinstance(composition, basestring):
            composition = parse_comp(composition)

        composition = defaultdict(float, composition)
        if self.bounds is None:
            return np.array([ composition[k] for k in self.sorted_space ])
        ctot = float(sum(composition.values()))
        cvec = np.array([ composition[k]/ctot for k in self.sorted_space ])
        coord = np.linalg.lstsq(self._basis.T, cvec)[0]
        if abs(sum(coord) - 1) > 1e-3 or any(c < -1e-3 for c in coord):
            raise PhaseSpaceError
        return coord

    def comp(self, coord):
        '''
        Returns the composition of a coordinate in phase space.

        Example:
        >>> space = PhaseSpace('Fe-Li-O')
        >>> space.comp([0.2, 0.2, 0.6])
        defaultdict(<type 'float'>, {'Fe': 0.20000000000000001, 
        'O':0.59999999999999998, 'Li': 0.20000000000000001})
        '''
        if self.bounds is None:
            return defaultdict(float, zip(self.sorted_space, coord))
        if len(coord) != len(self.bounds):
            raise PhaseSpaceError
        tot = sum(coord)
        coord = [ c/float(tot) for c in coord ]
        comp = defaultdict(float)
        for b, x in zip(self.bounds, coord):
            for elt, val in b.items():
                comp[elt] += val*x
        return defaultdict(float,
                dict( (k,v) for k,v in comp.items() if v > 0 ))

    def in_space(self, composition):
        '''
        Returns True, if the composition is in the right elemental-space 
        for this PhaseSpace.

        Example:
        >>> space = PhaseSpace('Fe-Li-O')
        >>> space.in_space('LiNiO2')
        False
        >>> space.in_space('Fe2O3')
        True
        '''

        if self.bounds is None:
            return True
        if isinstance(composition, Phase):
            composition = composition.comp
        elif isinstance(composition, basestring):
            composition = parse_comp(composition)

        if set(composition.keys()) <= self.space:
            return True
        else:
            return False

    def in_bounds(self, composition):
        '''
        Returns True, if the composition is within the bounds of the phase space

        Example:
        >>> space = PhaseSpace('Fe2O3-NiO2-Li2O')
        >>> space.in_bounds('Fe3O4')
        False
        >>> space.in_bounds('Li5FeO8')
        True
        '''
        if self.bounds is None:
            return True
        if isinstance(composition, Phase):
            composition = composition.unit_comp
        elif isinstance(composition, basestring):
            composition = parse_comp(composition)

        if not self.in_space(composition):
            return False

        try:
            self.coord(composition)
        except PhaseSpaceError:
            return False
        return True

    ### analysis stuff

    def get_qhull(self, phases=None):
        '''Get the convex hull for a given space.
        
        Sets the following variables:
            - stable
            - tie_lines
            - hull
        '''
        if phases is None: ## ensure there are phases to get the hull of
            phases = self.phase_dict.values()

        ## ensure that all phases have negative formation energies
        phases = [ p for p in phases if p.energy <= 0 ]

        phase_space = set()
        for phase in phases:
            phase_space |= phase.space

        A = []
        for phase in phases:
            A.append(list(self.coord(phase))[1:] + [phase.energy])

        dim = len(A[0])
        for i in range(dim):
            tmparr = [ 0 if a != i-1 else 1 for a in range(dim) ]
            if not tmparr in A:
                A.append(tmparr)

        A = np.array(A)
        if len(A) == len(A[0]):
            self._hull = set([frozenset([ p for p in phases])])
            self._tie_lines = set([ frozenset([k1, k2]) for k1, k2 in 
                    itertools.combinations(phases, r=2) ])
            self._stable = set([ p for p in phases])
            return

        #if len(A[0]) < len(phase_space):
        conv_hull = ConvexHull(A)
        #else:
        #    conv_hull = ConvexHull(A, joggle=True)

        hull = set()
        tie_lines = set()
        stable = set()
        for facet in conv_hull.simplices:
            ### various exclusion rules
            if any([ ind >= len(phases) for ind in facet ]):
                continue

            if all( phases[ind].energy == 0 for ind in facet 
                    if ind < len(phases)):
                continue

            dim = len(facet)
            face_matrix = np.array([ A[i] for i in facet ])
            face_matrix[:, -1] = 1
            v = np.linalg.det(face_matrix)
            
            if abs(v) < 1e-8:
                continue

            face = frozenset([ phases[ind] for ind in facet 
                if ind < len(phases)])

            stable |= set(face)
            tie_lines |= set([ frozenset([k1, k2]) for k1, k2 in
                    itertools.combinations(face, r=2)])
            hull.add(face)

        self._hull = hull 
        self._tie_lines = tie_lines 
        self._stable = stable

    def get_hull_points(self):
        '''
        Gets out-of PhaseSpace points. i.e. for FeSi2-Li, there are no other
        phases in the space, but there are combinations of Li-Si phases and
        Fe-Si phases. This method returns a list of phases including composite
        phases from out of the space.

        Example:
        >>> space = PhaseSpace('FeSi2-Li')
        >>> space.get_hull_points()
        [<Phase FeSi2 (23408): -0.45110217625>, 
        <Phase Li (104737): 0>, 
        <Phase 0.680 Li13Si4 + 0.320 FeSi : -0.3370691816>, 
        <Phase 0.647 Li8Si3 + 0.353 FeSi : -0.355992801765>, 
        <Phase 0.133 Fe3Si + 0.867 Li21Si5 : -0.239436904167>, 
        <Phase 0.278 FeSi + 0.722 Li21Si5 : -0.306877209723>]
        '''
        self._hull = set() # set of lists
        self._stable = set() # set
        done_list = [] # list of sorted lists
        hull_points = [] # list of phases

        if len(self.phases) == len(self.space):
            self._hull = set(frozenset(self.phases))
            self._stable = set(self.phases)
            return

        for b in self.bounds:
            e, x = self.gclp(b)
            p = Phase.from_phases(x)
            hull_points.append(p)
        
        facets = [list(hull_points)]
        while facets:
            facet = facets.pop(0)
            done_list.append(sorted(facet))
            phases, E = self.get_minima(self.phase_dict.values(), facet)
            phase = Phase.from_phases(phases)
            if not phase in hull_points:
                hull_points.append(phase)
                for new_facet in itertools.combinations(facet, r=len(facet)-1):
                    new_facet = list(new_facet) + [phase]
                    if new_facet not in done_list:
                        facets.append(new_facet)
        return hull_points

    def compute_meta_stability(self, phase):
        '''
        Compute the distance from the convex hull. For phases which are
        unstable this is simply the computed distance to the hull. For phases
        which are stable, we further compute the distance to the hull, without
        that phase. i.e. how stable is the next-most-favorable phase
        composition at that phase's composiion?
        '''
        if self.phase_dict[phase.name] != phase:
            stable = self.phase_dict[phase.name]
            phase.meta_stability = phase.energy - stable.energy
        elif len(phase.comp) == 1:
            stable = self.phase_dict[phase.name]
            phase.meta_stability = phase.energy - stable.energy
        else:
            phases = self.phase_dict.values()
            phases.remove(phase)
            energy, gclp_phases = self.gclp(phase.unit_comp, phases=phases)
            phase.meta_stability = phase.energy - energy
        #f = phase.formation
        #f.hull_distance = phase.meta_stability
        #f.save()

    def compute_meta_stabilities(self, phases=None):
        if phases is None:
            phases = self.phases
        for phase in phases:
            if phase.energy == 0:
                continue
            if not phase.meta_stability:
                self.compute_meta_stability(phase)

    ###
    ### Reaction stuff
    ###

    def get_reaction(self, var, facet=None):
        '''
        For a given composition, what is the maximum delta_composition reaction
        on the given facet. If None, returns the whole reaction for the given
        PhaseSpace.

        Example:
        >>> space = PhaseSpace('Fe2O3-Li2O')
        >>> equilibria = space.hull[0]
        >>> space.get_reaction('Li2O', facet=equilibria)
        '''

        if isinstance(var, basestring):
            var = parse_comp(var)

        if facet:
            phases = facet
        else:
            phases = self.stable

        prob = LpProblem('BalanceReaction', LpMaximize)
        pvars = LpVariable.dicts('prod', phases, 0)
        rvars = LpVariable.dicts('react', phases, 0)
        prob += sum([ p.fraction(var)['var']*pvars[p] for p in phases ])-\
                sum([ p.fraction(var)['var']*rvars[p] for p in phases ]),\
                "Maximize delta comp"
        for celt in self.space:
            prob += sum([ p.fraction(var)[celt]*pvars[p] for p in phases ]) ==\
                    sum([ p.fraction(var)[celt]*rvars[p] for p in phases ]),\
                    'identical %s composition on both sides' % celt
        prob += sum([ rvars[p] for p in phases ]) == 1
        prob.solve()
        #prob.solve(GUROBI(msg=False))
        prods = defaultdict(float,[ (c, pvars[c].varValue) for c in phases
            if pvars[c].varValue > 1e-6 ])
        reacts = defaultdict(float,[ (c, rvars[c].varValue) for c in phases
            if rvars[c].varValue > 1e-6 ])
        n_elt = value(prob.objective)
        return reacts, prods, n_elt

    def get_reactions(self, var, electrons=2.0):
        '''
        Returns a list of Reactions.

        Example:
        >>> space = PhaseSpace('Fe-Li-O')
        >>> space.get_reactions('Li', electrons=1)
        '''
        if isinstance(var, basestring):
            var = parse_comp(var)
        vname = format_comp(reduce_comp(v))
        vphase = self.phase_dict[vname]
        for facet in self.hull:
            reacts, prods, delta_var = self.get_reaction(var, facet=facet)
            if vphase in facet:
                yield Reaction(
                    products={vphase:sum(vphase.comp.values())}, 
                    reactants={}, delta_var=1.0,
                    electrons=electrons, variable=var)
                continue
            yield Reaction(products=prods, reactants=reacts, 
                delta_var=delta_var,
                variable=var, electrons=electrons)

    def plot_reactions(self, var, electrons=2.0, save=False):
        '''
        Plot the convex hull along the reaction path, as well as the voltage
        profile.
        '''
        if isinstance(var, basestring):
            var = parse_comp(var)
        vname = format_comp(var)

        fig = plt.figure()
        ax1 = fig.add_subplot(211)

        #plot tie lines
        for p1, p2 in self.tie_lines:
            c1 = self.coord(p1)
            c2 = self.coord(p2)
            if abs(c1[1]) < 1e-6 or abs(c2[1]) < 1e-6:
                if abs(c1[1] - 1) < 1e-6 or abs(c2[1] - 1) < 1e-6:
                    if len(self.tie_lines) > 1:
                        continue
            ax1.plot([c1[1],c2[1]], [p1.energy, p2.energy], 'k')

        #plot compounds
        for p in self.stable:
            x = self.coord(p.unit_comp)[1]
            ax1.plot(x, p.energy, 'bo')
            ax1.text(x, p.energy, '$\\rm{%s}$' % p.latex,
                    ha='left', va='top')
        plt.ylabel('$\\rm{\Delta H}$ $\\rm{[eV/atom]}$')
        ymin, ymax = ax1.get_ylim()
        ax1.set_ylim(ymin - 0.1, ymax)

        ax2 = fig.add_subplot(212, sharex=ax1)
        ax2.set_xlim([0,1])
        points = set()
        for reaction in self.get_reactions(var):
            voltage = reaction.delta_h/reaction.delta_var/electrons
            points |= set([(reaction.r_var_comp, voltage), 
                    (reaction.p_var_comp, voltage)])

        points = sorted( points, key=lambda x: x[0] )
        points = sorted( points, key=lambda x: -x[1] )
        base = sorted(self.stable, key=lambda x:
                x.fraction(var)['var'])[0]

        if len(points) > 1:
            for i in range(len(points) - 2):
                ax2.plot([points[i][0], points[i+1][0]],
                        [points[i][1], points[i+1][1]], 'k')

            ax2.plot([points[-2][0], points[-2][0]], 
                    [points[-2][1], points[-1][1]], 'k')
            ax2.plot([points[-2][0], 1], 
                    [points[-1][1], points[-1][1]], 'k')
        else:
            ax2.plot([0, 1], [points[0][1], points[0][1]], 'k')
        
        plt.xlabel('$\\rm{x}$ $\\rm{in}$ $\\rm{(%s)_{x}(%s)_{1-x}}$' % ( 
            comp_to_latex(var),
            base.latex))
        plt.ylabel('$\\rm{Voltage}$ $\\rm{[V]}$')

        return ax1, ax2

        #if not save:
        #    plt.show()
        #else:
        #    plt.savefig('%s-%s.eps' % (save, vname),
        #            bbox_inches='tight', 
        #            transparent=True,
        #            pad_inches=0) 



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import warnings
import fractions as frac
import numpy as np

from django.db import models

from routines import *

class TranslationError(Exception):
    pass

class RotationError(Exception):
    pass

class OperationError(Exception):
    pass

class WyckoffSiteError(Exception):
    pass

class SpacegroupError(Exception):
    pass

class Translation(models.Model):
    x = models.FloatField()
    y = models.FloatField()
    z = models.FloatField()

    class Meta:
        app_label = 'qmpy'
        db_table = 'translations'

    @property
    def vector(self):
        return np.array([self.x, self.y, self.z])

    @vector.setter
    def vector(self, vector):
        self.x, self.y, self.z = vector

    @classmethod
    def get(cls, vector):
        fields = ['x', 'y', 'z']
        arr_dict = dict(zip(fields, vector))
        obj, new = cls.objects.get_or_create(**arr_dict)
        if new:
            obj.save()
        return obj

    def __str__(self):
        ops = []
        for t in self.vector:
            if t == 0:
                s = '0'
            elif t < 0:
                s = '-%s' % (frac.Fraction(str(t)))
            else:
                s = '+%s' % (frac.Fraction(str(t)))
            ops.append(s)
        return ','.join(ops)

class Rotation(models.Model):
    a11 = models.FloatField()
    a12 = models.FloatField()
    a13 = models.FloatField()
    a21 = models.FloatField()
    a22 = models.FloatField()
    a23 = models.FloatField()
    a31 = models.FloatField()
    a32 = models.FloatField()
    a33 = models.FloatField()

    class Meta:
        app_label = 'qmpy'
        db_table = 'rotations'

    @property 
    def matrix(self):
        return np.array([ 
            [ self.a11, self.a12, self.a13 ],
            [ self.a21, self.a22, self.a23 ],
            [ self.a31, self.a32, self.a33 ]])

    @matrix.setter
    def matrix(self, matrix):
        self.a11, self.a12, self.a13 = matrix[0]
        self.a21, self.a22, self.a23 = matrix[1]
        self.a31, self.a32, self.a33 = matrix[2]

    @classmethod
    def get(cls, matrix):
        fields = [ 'a11', 'a12', 'a13',
                   'a21', 'a22', 'a23',
                   'a31', 'a32', 'a33']
        matrix = np.ravel(matrix)
        mat_dict = dict(zip(fields, matrix))
        obj, new = cls.objects.get_or_create(**mat_dict)
        if new:
            obj.save()
        return obj

    def __str__(self):
        ops = []
        indict = {0:'x', 1:'y', 2:'z'}
        for r in self.matrix:
            s = ''
            for i, x in enumerate(r):
                if x == 0:
                    continue
                elif x == 1:
                    s += indict[i]
                elif x == -1:
                    s += '-'+indict[i]
                else:
                    f = frac.Fraction(str(x))
                    s += '%s%s' % (f, indict[i])
            ops.append(s)
        return ','.join(ops)

class Operation(models.Model):
    rotation = models.ForeignKey(Rotation)
    translation = models.ForeignKey(Translation)

    class Meta:
        app_label = 'qmpy'
        db_table = 'operations'

    @classmethod
    def get(cls, value):
        '''
        Accepts symmetry operation strings, i.e. "+x, x+1/2, x+y-z" or a tuple
        of rotation matrix and translation vector. 

        Example:
        >>> Operation.get("x,y,-y")
        >>> Operation.get(( rot, trans ))
        '''
        if isinstance(value, basestring):
            rot, trans = parse_sitesym(value)
        elif isinstance(value, tuple):
            rot, trans = value
        rot = Rotation.get(rot)
        trans = Translation.get(trans)
        op, new = cls.objects.get_or_create(rotation=rot, translation=trans)
        if new:
            op.save()
        return op

    def __str__(self):
        ops = []
        indict = {0:'x', 1:'y', 2:'z'}
        for r,t in zip(self.rotation.matrix, 
                       self.translation.vector):

            s = ''
            for i, x in enumerate(r):
                if x == 0:
                    continue
                elif x == 1:
                    s += indict[i]
                elif x == -1:
                    s += '-'+indict[i]
                else:
                    f = frac.Fraction(str(x))
                    s += '%s%s' % (f, indict[i])

            if t == 0:
                pass
            elif t < 0:
                s += '-%s' % (frac.Fraction('%08f' % t))
            else:
                s += '+%s' % (frac.Fraction('%08f' % t))
            ops.append(s)
        return ','.join(ops)


class WyckoffSite(models.Model):
    spacegroup = models.ForeignKey('Spacegroup', related_name='site_set')
    symbol = models.CharField(max_length=1)
    multiplicity = models.IntegerField(blank=True, null=True)
    x = models.CharField(max_length=8)
    y = models.CharField(max_length=8)
    z = models.CharField(max_length=8)
    
    class Meta:
        app_label = 'qmpy'
        db_table = 'wyckoffsites'

    def __str__(self):
        return '%s%d' % (self.symbol, self.multiplicity)

    @classmethod
    def get(cls, symbol, spacegroup):
        site, new = cls.objects.get_or_create(spacegroup=spacegroup, 
                symbol=symbol)
        if new:
            site.save()
        return site

class Spacegroup(models.Model):
    number = models.IntegerField(primary_key=True)
    hm = models.CharField(max_length=30, blank=True, null=True)
    hall = models.CharField(max_length=30, blank=True, null=True)
    pearson = models.CharField(max_length=30)
    schoenflies = models.CharField(max_length=30)
    operations = models.ManyToManyField(Operation, null=True)
    centering_vectors = models.ManyToManyField(Translation)
    lattice_system = models.CharField(max_length=20)
    centrosymmetric = models.BooleanField()

    _sym_ops = None
    _rots = None
    _trans = None

    class Meta:
        app_label = 'qmpy'
        db_table = 'spacegroups'

    def save(self, *args, **kwargs):
        super(Spacegroup, self).save(*args, **kwargs)
        for op in self.sym_ops:
            op.save()
        self.operations = self.sym_ops

    @property
    def sym_ops(self):
        if self._sym_ops is None:
            self._sym_ops = [ op for op in self.operations.all() ]
        return self._sym_ops

    @sym_ops.setter
    def sym_ops(self, sym_ops):
        self._sym_ops = sym_ops

    @property
    def rotations(self):
        if self._rots is None:
            self._rots = np.array([ op.rotation.matrix for op in self.sym_ops ])
        return self._rots

    @rotations.setter
    def rotations(self, rotations):
        self._rots = rotations

    @property
    def translations(self):
        if self._trans is None:
            self._trans = np.array([ op.translation.array for op in self.sym_ops ])
        return self._trans

    @translations.setter
    def translations(self, translations):
        self._trans = translations

    def __str__(self):
        return str(self.number)

    @property
    def wyckoff_sites(self):
        return self.site_set.all().order_by('symbol')

    @property
    def symbol(self):
        return self.hm

    def equivalent_sites(self, structure, symprec=1e-3):
        sites = []
        for atom in structure.uniq_atoms:
            for rot, trans in zip(self.rotations, self.translations):
                site = np.dot(rot, atom.coord) + trans
                site %= 1.0
                a = atom.copy()
                a.coord = site
                if not any( s == a for s in sites ):
                    sites.append(a)
        return sites

    def symmetry_normalised_sites(self, scaled_positions):
        """Returns an array of same size as *scaled_positions*,
        containing the corresponding symmetry-equivalent sites within
        the unit cell of lowest indices.

        Example:

        >>> from ase.lattice.spacegroup import Spacegroup
        >>> sg = Spacegroup(225)  # fcc
        >>> sg.symmetry_normalised_sites([[0.0, 0.5, 0.5], [1.0, 1.0, 0.0]])
        array([[ 0.,  0.,  0.],
               [ 0.,  0.,  0.]])
        """
        scaled = np.array(scaled_positions, ndmin=2)
        normalised = np.empty(scaled.shape, np.float)
        for i, pos in enumerate(scaled):
            sympos = np.dot(self.rotations, pos) + self.translations
            sympos %= 1.0
            sympos %= 1.0
            j = np.lexsort(sympos.T)[0]
            normalised[i,:] = sympos[j]
        return normalised

    def unique_sites(self, structure, symprec=1e-3, output_mask=False):
        scaled = np.array(structure.scaled_coords, ndmin=2)
        symnorm = self.symmetry_normalised_sites(scaled)
        perm = np.lexsort(symnorm.T)
        iperm = perm.argsort()
        xmask = np.abs(np.diff(symnorm[perm], axis=0)).max(axis=1) > symprec
        mask = np.concatenate(([True], xmask))
        imask = mask[iperm]
        if output_mask:
            return [ a for a, i in zip(structure.atoms, imask) if i ], imask
        else:
            return [ a for a, i in zip(structure.atoms, imask) if i ]

    @property
    def n_sym_ops(self):
        return self.operations.count()

    @property
    def n_wyckoff_sites(self):
        return self.site_set.count()

    def get_site(self, symbol):
        symol = symbol.strip('0123456789')
        return self.site_set.get(symbol__exact=symbol)



import numpy as np
import numpy.linalg as linalg
import itertools
import matplotlib
matplotlib.rcParams.update({'font.size': 22,
    'font.serif':['Century']})
import matplotlib.pylab as plt
from collections import defaultdict
from StringIO import StringIO

def get_pair_distances(structure, max_dist=10):
    dists = []
    structure.symmetrize()
    cell = structure.cell
    coords = structure.cartesian_coords.copy()
    lat_params = structure.lat_params
    limits = [ int(np.ceil(max_dist/lat_params[i])) for i in range(3) ]
    ranges = [ range(-l, l+1) for l in limits ]
    sc = np.vstack([ coords+np.dot(ijk, cell) for ijk in itertools.product(*ranges)])
    elts = [ a.element_id for a in structure ]
    n_cells = len(ranges[0])*len(ranges[1])*len(ranges[2])
    elt_list = list(set(elts))
    elts = n_cells*elts

    pairs = list(itertools.combinations(elt_list, r=2))
    dists = defaultdict(list)
        
    for a1 in structure.uniq_sites:
        c1 = a1.cart_coord
        e1 = a1.label
        for c2, e2 in zip(sc, elts):
            vec = c1 - c2
            if all([ v == 0 for v in vec ]):
                continue
            dists[frozenset([e1,e2])] += [linalg.norm(vec)]*a1.multiplicity
    return dists

def get_pdf_uri(structure, max_dist=5, smearing=0.1):
    distances = get_pair_distances(structure, max_dist=max_dist)
    elts = list(set([ a.element_id for a in structure ]))
    rhos = dict((elt, structure.comp[elt]/structure.volume) for elt in elts)

    xs = np.mgrid[0:max_dist:200j]
    dr = xs[1] - xs[0]
    ideals = dict((elt, 4.0*np.pi*rhos[elt]/3.0) for elt in elts)

    for pair, dists in distances.items():
        pair = list(pair)
        if len(pair) == 1:
            pair *= 2
        e1, e2 = pair

        vals = np.zeros(xs.shape)
        prefactor = 1.0/(smearing*np.sqrt(2*np.pi))
        norms = [ ( (x+dr/2)**3 - (x-dr/2)**3) for x in xs ]
        for d in dists:
            vals += np.exp(-(d-xs)**2/(2*smearing**2))
            #vals += prefactor*np.exp(-(d-xs)**2/(2*smearing**2))
        vals = vals/norms
        plt.plot(xs, vals, lw=3, label="%s-%s" % (e1, e2))

    plt.legend()
    plt.xlabel('interatomic distance [$\\AA$]')
    img = StringIO()
    plt.savefig(img, dpi=75, bbox_inches='tight')
    data_uri = 'data:image/jpg;base64,'
    data_uri += img.getvalue().encode('base64').replace('\n', '')
    plt.close()
    return data_uri



#!/usr/bin/env python

#= Represent materials, and their attributes =#
from materials.entry import *
from materials.structure import *
from materials.composition import *
from materials.element import *
from materials.xrd import *

#= DFT Calculations and computed properties =#
from analysis.vasp.calculation import *
from analysis.vasp.dos import *
from analysis.vasp.potential import *

#= Local resources and computing =#
from computing.queue import *
from computing.resources import *
from computing.scripts import *

#= Other analyses =#
from analysis import *
from analysis.symmetry import *

#= Known experimental data, and meta data =#
from data.meta_data import *
from data.experiment import *

#= Custom database models and fields =#
from qmpy.db.custom import *



'''
Implementation of the Bond-Valence Sum method.

Data sourced from: 
    http://www.iucr.org/resources/data/datasets/bond-valence-parameters
'''

import numpy as np

#from qmpy.data.bond_valence import bond_valence_parameters
#from qmpy.data.bond_valence import bond_valence_references

def param_getter(symbol1, symbol2):
    pair = set([ symbol1, symbol2])
    #results = []
    for data in bond_valence_parameters:
        if set([data['A_species'], data['B_species']]) == pair:
            #results.append(data)
            return data
    #return results

def v_ij(atom1, atom2, params=None):
    if params is None:
        params = param_getter(atom1.element_id, atom2.element_id)
    if params is None:
        return None
    R = atom1.structure.get_distance(atom1, atom2)
    R0 = params['param_r0']
    B = params['param_B']
    return np.exp( (R0 - R)/B )

def valence_sum(atom):
    valence = 0
    structure = atom.structure
    for atomp in structure.atoms:
        if atomp is atom:
            pass
        elif structure.get_distance(atom, atomp) < 2.5:
            v = v_ij(atom, atomp)
            if v is None:
                continue
            valence += v
    return valence

def total_valence_sum(structure):
    for atom in structure.atoms:
        atom.charge = valence_sum(atom)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import numpy.linalg as linalg
from scipy.spatial import Delaunay, Voronoi
import matplotlib.pylab as plt
from mpl_toolkits.mplot3d import Axes3D
import itertools
from qmpy.utils import *

def get_facet_area(vertices):
    area = 0
    vertices = np.sort(vertices)
    #vertices = vertices[np.lexsort(vertices.T)]
    if not vertices.shape[1] == 3:
        raise ValueError('vertices must be an Nx3 array')
    for i in range(len(vertices)-2):
        a = vertices[i]
        b = vertices[i+1]
        c = vertices[i+2]
        ab = a-b
        ac = a-c
        area += np.dot(ab, ac)
    return area

def find_neighbors(structure):
    structure.make_conventional()
    cell = structure.cell
    scaled = structure.scaled_coords
    x8 = np.vstack([ structure.scaled_coords + 
                     i*np.array([1, 0, 0]) + 
                     j*np.array([0, 1, 0]) + 
                     k*np.array([0, 0, 1])
            for i, j, k in itertools.product([0, -1, 1], 
                                             [0, -1, 1], 
                                             [0, -1, 1])])
    nns = {}
    for i, atom in enumerate(structure.atoms):
        # recenter cell
        #coords = x8 - atom.coord
        #coords[ coords < -2.5 ] += 5
        #coords[ coords > 2.5 ] -= 5
        carts = np.dot(x8, cell)
        tess = Voronoi(carts)
        n = 0
        if atom.element_id != 'Al':
            continue
        for j, r in enumerate(tess.ridge_points):
            if not i in r:
                continue
            inds = tess.ridge_vertices[j]
            verts = tess.vertices[inds]
            area = get_facet_area(verts)
            if abs(area) <  5e-2:
                continue
            n += 1
        print '%s nearest neighbors found!' % n
    #for i in range(len(scaled)):
        # recenter cell
        #carts = np.dot(x27 - scaled[i], cell)
        #mask = np.argwhere([(linalg.norm(x) < 3 and not all(x == [0,0,0]))
        #                      for x in carts])
        #potential = np.vstack(carts[mask])
        #print potential
        #kdtree = KDTree(potential)
        #return kdtree

def get_pdf(structure, ind=None):
    result = []
    for atom1 in structure:
        dists = []
        for atom2 in structure:
            dists.append(structure.get_distance(atom1, atom2))
        result.append(dists)
    kernel = gaussian_kde(result[0])
    kernel.covariance_factor = lambda : .25
    kernel._compute_covariance()
    xs = np.linspace(1, max(result[0]), 1000)
    plt.plot(xs, kernel(xs))
    plt.hist(result[0], normed=True)
    plt.show()
    return result



import logging
import time
from django.db import models
from datetime import datetime
import os

import qmpy.utils as utils
from qmpy.analysis.vasp import *

logger = logging.getLogger(__name__)

def initialize(entry,*args, **kwargs):
    entry.input.set_magnetism('ferro')
    calc = Calculation.setup(entry.input, entry=entry,
            configuration='initialize', path=entry.path+'/initialize', 
            kwargs=kwargs)
    if not calc.converged:
        calc.write()
        return calc

    if calc.converged:
        if calc.magmom > 0.1:
            entry.keywords.append('magnetic')
        entry.calculations['initialize'] = calc
        calc.output = calc.input
    return calc

def coarse_relax(entry, *args, **kwargs):
    if entry.calculations.get('coarse_relax', Calculation()).converged:
        return entry.calculations['coarse_relax']

    calc = initialize(entry, **kwargs)
    if not calc.converged:
        calc.write()
        return calc

    inp = entry.structures['input']
    if not 'magnetic' in entry.keywords:
        inp.set_magnetism('none')
    calc = Calculation.setup(inp, entry=entry,
        configuration='coarse_relax', path=entry.path+'/coarse_relax',
        **kwargs)

    if calc.converged:
        entry.structures['coarse_relax'] = calc.output
        entry.calculations['coarse_relax'] = calc
    return calc

def fine_relax(entry, *args, **kwargs):
    if entry.calculations.get('fine_relax', Calculation()).converged:
        return entry.calculations['fine_relax']

    calc = coarse_relax(entry, **kwargs)
    if not calc.converged:
        calc.write()
        return calc
    inp = entry.structures['coarse_relax']
    if 'magnetic' in entry.keywords:
        inp.set_magnetism('ferro')

    calc = Calculation.setup(inp, entry=entry,
        configuration='fine_relax', path=entry.path+'/fine_relax',
        *args, **kwargs)
    if calc.converged:
        entry.structures['fine_relax'] = calc.output
        entry.calculations['fine_relax'] = calc
    return calc

def standard(entry, *args, **kwargs):
    if entry.calculations.get('standard', Calculation()).converged:
        return entry.calculations['standard']

    calc = fine_relax(entry, **kwargs)
    if not calc.converged:
        calc.write()
        return calc

    inp = entry.structures['fine_relax']
    if 'magnetic' in entry.keywords:
        inp.set_magnetism('ferro')

    calc = Calculation.setup(inp, entry=entry,
        configuration='standard', path=entry.path+'/standard', **kwargs)
    if calc.converged:
        calc.output = calc.input
        calc.compute_formation()
        entry.calculations['standard'] = calc
    return calc

def relaxation(entry, *args, **kwargs):
    if entry.calculations.get('relaxation', Calculation()).converged:
        return entry.calculations['relaxation']

    calc = Calculation.setup(entry.input, *args, entry=entry,
                                          configuration='relaxation',
                                          path=entry.path+'/relaxation',
                                          **kwargs)
    entry.calculations['relaxation'] = calc
    if not calc.converged:
        calc.write()
        return calc
    else:
        entry.structures['relaxation'] = calc.output
    return calc

def static(entry, *args, **kwargs):
    if entry.calculations.get('static', Calculation()).converged:
        return entry.calculations['static']

    calc = relaxation(entry, *args, **kwargs)
    if not calc.converged:
        calc.write()
        return calc

    input = calc.output
    calc = Calculation.setup(input, entry=entry,
                                    configuration='static', 
                                    path=entry.path+'/static', 
                                    **kwargs)
    entry.calculations['static'] = calc

    if calc.converged:
        calc.output = calc.input
        calc.compute_formation()
        entry.calculations['static'] = calc
    return calc



#!/usr/bin/env python

from numpy import *
from collections import defaultdict
import os.path
import time
import qmpy
import StringIO

from qmpy.utils import *

THERMOPY_LIB_PATH = qmpy.INSTALL_PATH+'/data/thermodynamic/'

class PhaseError(Exception):
    pass

class PhaseDataError(Exception):
    pass

class PhaseData(object):
    def __init__(self):
        self._phases = []
        self._phase_dict = {}

    def __eq__(self, other):
        if self.space != other.space:
            return False
        if any( abs(self.unit_comp[k] - other.unit_comp[k]) > 1e-6 for k in
                self.unit_comp.keys() ):
            return False
        if abs( self.energy - other.energy ) > 1e-6:
            return False

    @property
    def phases(self):
        '''
        Return all phases.
        '''
        return self._phases

    @phases.setter
    def phases(self, phases):
        self._phases = phases
        self._phase_dict = None

    @property
    def phase_dict(self):
        if not self._phase_dict:
            self.get_phase_dict()
        return self._phase_dict

    @phase_dict.setter
    def phase_dict(self, phase_dict):
        self._phase_dict = phase_dict

    @property
    def space(self):
        space = set()
        for p in self.phases:
            space |= p.space
        return space

    ### Methods

    def get_phase_dict(self):
        '''
        Dictionary of phase.name:phase, such that each composition is mapped to
        the lowest energy phase at that composition.

        Example:
        >>> pd = PhaseData()
        >>> pd.add_phase(Phase(composition='Fe2O3', energy=-3))
        >>> pd.add_phase(Phase(composition='Fe2O3', energy=-4))
        >>> pd.add_phase(Phase(composition='Fe2O3', energy=-5))
        >>> pd.phase_dict
        {'Fe2O3': <Phase Fe2O3 : -5}
        '''
        phase_dict = {}
        for p in self.phases:
            if not p.name in phase_dict:
                phase_dict[p.name] = p
            else:
                if p.energy < phase_dict[p.name].energy:
                    phase_dict[p.name] = p
        self.phase_dict = phase_dict

    def add_phase(self, phase):
        '''
        Add a phase, updated the phase dictionary appropriately.
        '''
        self.phases.append(phase)
        if not phase.name in self.phase_dict:
            self.phase_dict[phase.name] = phase
        else:
            if phase.energy < self.phase_dict[phase.name].energy:
                self.phase_dict[phase.name] = phase

    def add_phases(self, phases):
        for phase in phases:
            self.add_phase(phase)

    def load_library(self, library):
        '''
        Load a library file, containing self-consistent thermochemical data.
        
        '''
        self.read_file(qmpy.INSTALL_PATH+'/data/thermodata/'+library)

    def load_oqmd(self, space=None, search={}, stable=False):
        from qmpy.analysis.vasp.calculation import Calculation
        from qmpy.materials.element import Element
        data = Calculation.objects.exclude(delta_e=None)
        #if stable:
        #    data = data.exclude(stability__lte=0)
        if search:
            data = data.filter(**search)
        if space:
            space_qs = Element.objects.exclude(symbol__in=space)
            data = data.filter(composition__element_set__in=space)
            data = data.exclude(composition__element_set__in=space_qs)

        for p in data:
            self.add_phase(p.phase)

    def read_file(self, filename, per_atom=True):
        '''
        Read in a thermodata file (named filename), formatted as 
        "composition energy [description]"

        E.g.

        composition energy
        Fe 0.0
        O 0.0
        Li 0.0
        Fe3O4 -0.21331204979
        FeO -0.589343204057
        Fe3O4 -0.21331204979
        FeLiO2 -0.446739168889
        FeLi5O4 -0.198830531099
        '''
        if isinstance(filename, basestring):
            fileobj = open(filename)
        elif isinstance(filename, file):
            fileobj = filename
        elif isinstance(filename, type(StringIO.StringIO())):
            fileobj = filename
            fileobj.name = None
        thermodata = fileobj.readlines()
        headers = [ h.lower() for h in thermodata.pop(0).strip().split() ]
        if 'composition' not in headers:
            raise PhaseDataError("Found columns: %s. Must provide composition in\
                                  a column labelled composition." % 
                                  (', '.join(headers)))
        if ('energy' not in headers and 'delta_e' not in headers):
            raise PhaseDataError("Found columns: %s. Must provide energies in\
                                  a column labelled delta_e or energy." % 
                                  (', '.join(headers)))

        keywords = {'energy':'energy', 'composition':'composition', 
                'delta_e':'energy', 'delta_h':'energy', 'delta_g':'energy',
                'comp':'composition', 'name':'composition', 
                'desc':'description', 'description':'description'}

        headers = [ keywords[h] for h in headers if h in keywords ]

        for i, line in enumerate(thermodata):
            line = line.strip().split()
            if not line:
                continue
            ddict = dict(zip(headers, line))
            phase = Phase(composition=ddict['composition'],
                          energy=float(ddict['energy']),
                          description=ddict.get('description', 
                              '{file}:{line}'.format(file=fileobj.name, line=i)),
                          per_atom=per_atom)
            self.add_phase(phase)

class Phase(object):
    '''
    A Phase represents a computed formation energy, and its corresponding
    composition. Phases will also have description which indicates something
    about its source. If loaded from OQMD the Formation object imported can be
    accessed by Phase._formation and the formation id can be accessed
    by Phase.id.
    '''

    def __init__(self, 
            composition=None, 
            energy=None, 
            description='',
            per_atom=True,
            name=''):

        if composition is None or energy is None:
            raise IncompletePhaseError
        if isinstance(composition, basestring):
            composition = parse_comp(composition)

        self._unit_comp = None
        self._nom_comp = None
        self._space = None
        self._latex = None
        self._meta_stability = None
        self._name = name
        self._description = description
        self._comp = defaultdict(float, composition)
        self._formation = None
        self._phase_dict = None
        self._gap = None
        self.id = None

        if not per_atom:
            self._total_energy = energy
            self._energy = energy/sum(composition.values())
            self._energy_pfu = energy / sum(self.nom_comp.values())
        else:
            self._energy = energy
            self._total_energy = energy * sum(composition.values())
            self._energy_pfu = energy / sum(self.nom_comp.values())

        if len(composition) == 1:
            self._energy = 0
            self._total_energy = 0
            self._energy_pfu = 0

    @staticmethod
    def from_phases(phase_dict):
        '''
        Generate a Phase object from a dictionary of Phase objects. Returns a
        composite phase of unit composition.
        '''
        if len(phase_dict) == 1:
            return phase_dict.keys()[0]

        pkeys = sorted(phase_dict.keys(), key=lambda x: x.name)
        energy = sum([ amt*p.energy for p, amt in phase_dict.items() ])
        name = ''
        latex = ''
        for p in pkeys:
            amt = phase_dict[p]
            if amt < 1e-3:
                continue
            if amt == 1:
                name += p.name
                latex += p.latex
            else:
                name += '%s %s' % (round(amt,3), p.name)
                latex += '%s %s' % (round(amt,3), p.latex)
            name += ' + '
            latex += ' + '
        name = name.rstrip(' + ')
        latex = latex.rstrip(' + ')

        comp = defaultdict(float)
        for p, factor in phase_dict.items():
            for e, amt in p.unit_comp.items():
                comp[e] += amt*factor

        phase = Phase(
                composition=comp,
                energy=energy,
                per_atom=False)
        phase._name = name
        phase._latex = latex
        phase._phase_dict = phase_dict
        return phase

    def __repr__(self):
        if self._description:
            return '<Phase {name} ({description}): {energy}>'.format(
                    name=self.name, energy=self.energy, description=self._description)
        else:
            return '<Phase {name} : {energy}>'.format(
                    name=self.name, energy=self.energy)

    def __eq__(self,other):
        '''
        Phases are defined to be equal if they have the same composition and an
        energy within 1e-6 eV/atom.
        '''
        if set(self.comp) != set(other.comp):
            return False
        if abs(self.energy - other.energy) > 1e-6:
            return False
        for key in self.comp:
            if abs(self.unit_comp[key]-other.unit_comp[key]) > 1e-6:
                return False
        return True

    #### Attributes

    @property
    def name(self):
        '''
        Alphabetical ordering of composition.
        '''
        if not self._name:
            self._name = format_comp(self.comp, special='reduce')
        return self._name

    @property
    def volume(self):
        if self._phase_dict is not None:
            return sum( phase.formation.entry.volume*amt for phase, amt in
                    self._phase_dict.items() )
        else:
            return self.formation.entry.volume

    @property
    def mass(self):
        if self._phase_dict is not None:
            return sum( phase.formation.composition.mass*amt for phase, amt in
                    self._phase_dict.items() )
        else:
            return self.formation.composition.mass

    @property
    def space(self):
        '''
        Set of elements in the phase.
        '''
        return set([ k for k, v in self.unit_comp.items() 
            if abs(v) > 1e-6 ])

    @property
    def n(self):
        '''
        Number of atoms in the total composition.
        '''
        return sum(self._comp.values())

    @property
    def name(self):
        return format_comp(self.comp)

    @property
    def comp(self):
        '''
        Total composition.
        '''
        return self._comp

    @property
    def unit_comp(self):
        '''
        Unit composition.
        '''
        return unit_comp(self.comp)

    @property
    def nom_comp(self):
        '''
        Composition divided by the GCD. e.g. Fe4O6 becomes Fe2O3.
        '''
        return reduce_comp(self.comp)

    @property
    def latex(self):
        '''
        LaTeX friendly format of name.
        '''
        return latex_comp(self.nom_comp)

    @property
    def name(self):
        return format_comp(reduce_comp(self.comp))

    @property
    def description(self):
        '''
        Some description of the phase origin.
        '''
        return self._description

    @property
    def energy(self):
        '''
        Energy per atom in eV.
        '''
        return self._energy

    @property
    def total_energy(self):
        '''
        Total energy for the composition as supplied (in eV).
        '''
        return self._total_energy

    @property
    def energy_pfu(self):
        '''
        Energy per nominal composition. i.e. energy per Fe2O3, not Fe4O6.
        '''
        return self._energy_pfu

    @property
    def formation(self):
        '''
        OQMD Formation object, if it exists.
        '''
        if self._formation is None:
            self.get_formation()
        return self._formation

    @property
    def band_gap(self):
        if not self._gap:
            self.get_gap()
        return self._gap

    def get_gap(self):
        if not self._phase_dict:
            self._gap = self.formation.calculation.band_gap
        else:
            self._gap = min([ p.formation.calculation.band_gap for p in
                self._phase_dict ])

    #### Methods

    @property
    def calculation(self):
        '''
        Get the oqmd Formation object for this Phase, if it exists.
        '''
        if self.id is None:
            return
        from qmpy.analysis.vasp.calculation import Calculation
        return Calculation.objects.get(id=self.id)

    def free_energy(self, T=0, P=0, mus={}):
        '''
        Free energy function for the phase, can be defined to be anything, by
        default it just returns the phase's ground state energy.
        '''
        #global environment
        return self._energy

    def fraction(self, comp):
        '''
        Returns a composition dictionary with the specified composition pulled
        out as 'var'. 

        Example:
        >>> phase = Phase(composition={'Fe':1, 'Li':5, 'O'8'}, energy=-1)
        >>> phase.fraction('Li2O')
        defaultdict(<type 'float'>, {'var': 2.5, 'Fe': 1, 'O': 5.5, 'Li': 0.0})
        '''
        if isinstance(comp, Phase):
            comp = comp.unit_comp
        elif isinstance(comp, basestring):
            comp = parse_comp(comp, special='unit')
        residual = defaultdict(float, self.unit_comp)
        tot = sum(residual.values())
        for c, amt in dict(comp).items():
            pres = residual[c]/amt
            for c2, amt2 in comp.items():
                residual[c2] -= pres*amt2
        residual['var'] = (tot - sum(residual.values()))
        residual['var'] /= float(sum(comp.values()))
        return residual



'''
Useful symmetry functions.
'''

import qmpy.utils._spglib as spg
from qmpy.data import elements
import fractions as frac
import numpy as np

## spglib functions | http://spglib.sourceforge.net/

zdict = dict( (e['z'], e['symbol']) for e in elements.values() )

def find_structure_symmetry(structure, method='spglib',
        symprec=1e-5, angle_tolerance=-1.0):
    """
    Return the rotatiosn and translations which are possessed by the structure.
    
    Example:
    >>> from qmpy.io import read
    >>> from qmpy.analysis.symmetry import find_structure_symmetry
    >>> structure = read('POSCAR')
    >>> find_structure_symmetry(structure)
    
    """
    # Get number of symmetry operations and allocate symmetry operations
    multi = 48 * len(structure)
    rotation = np.zeros((multi, 3, 3), dtype=int)
    translation = np.zeros((multi, 3))
  
    # Get symmetry operations
    magmoms = structure.magmoms
    if not any(magmoms):
        num_sym = spg.symmetry(rotation,
                               translation,
                               structure.cell.T.copy(),
                               structure.scaled_coords.copy(),
                               structure.atomic_numbers,
                               symprec,
                               angle_tolerance)
    else:
        num_sym = spg.symmetry_with_collinear_spin(rotation,
                                                   translation,
                                                   lattice,
                                                   positions,
                                                   numbers,
                                                   magmoms,
                                                   symprec,
                                                   angle_tolerance)
  
    return rotation[:num_sym], translation[:num_sym]

def get_symmetry_dataset(structure, symprec=1e-3, angle_tolerance=-1.0):
    """
    Return a full set of symmetry information from a given input structure.

    Mapping values:
        number: International space group number
        international: International symbol
        hall: Hall symbol
        transformation_matrix:
          Transformation matrix from lattice of input cell to Bravais lattice
          L^bravais = L^original * Tmat
        origin shift: Origin shift in the setting of 'Bravais lattice'
        rotations, translations:
          Rotation matrices and translation vectors
          Space group operations are obtained by
            [(r,t) for r, t in zip(rotations, translations)]
        wyckoffs:
          Wyckoff letters

    Example:
    >>> from qmpy.io import read
    >>> from qmpy.analysis.symmetry import get_symmetry_dataset
    >>> structure = read('POSCAR')
    >>> get_symmetry_dataset(structure)

    """
    keys = ('number',
            'international',
            'hall',
            'transformation_matrix',
            'origin_shift',
            'rotations',
            'translations',
            'wyckoffs',
            'equivalent_atoms')

    scaled = structure.site_coords.copy()
    labels = structure.site_labels.copy()
    ldict = dict( (k, i+1) for i, k in enumerate(list(set(labels))))
    numbers = np.array([ ldict[k] for k in labels ])

    dataset = {}
    for key, data in zip(keys, 
            spg.dataset(
                structure.cell.T.copy(),
                scaled,
                numbers,
                symprec,
                angle_tolerance)):
        dataset[key] = data

    dataset['international'] = dataset['international'].strip()
    dataset['hall'] = dataset['hall'].strip()
    dataset['transformation_matrix'] = np.array(dataset['transformation_matrix']).T
    dataset['origin_shift'] = np.array(dataset['origin_shift'])
    dataset['rotations'] = np.array(dataset['rotations'])
    dataset['translations'] = np.array(dataset['translations'])
    letters = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
    dataset['wyckoffs'] = [letters[x] for x in dataset['wyckoffs']]
    dataset['equivalent_atoms'] = np.array(dataset['equivalent_atoms'])

    return dataset

def get_spacegroup(structure, symprec=1e-5, angle_tolerance=-1.0):
    """
    Return space group in international table symbol and number
    as a string.
    """
    # Atomic positions have to be specified by scaled positions for spglib.
    return int(spg.spacegroup(structure.cell.T.copy(),
                          structure.scaled_coords.copy(),
                          structure.atomic_numbers,
                          symprec,
                          angle_tolerance).strip(' ()'))

def get_pointgroup(rotations):
    """
    Return point group in international table symbol and number.
    """

    # (symbol, pointgroup_number, transformation_matrix)
    return spg.pointgroup(rotations)

def refine_cell(structure, symprec=1e-5, angle_tolerance=-1.0):
    """
    Return refined cell
    """
    # Atomic positions have to be specified by scaled positions for spglib.
    num_atom = len(structure)
    lattice = structure.cell.T.copy()
    positions = np.zeros((num_atom * 4, 3), dtype=float)
    positions[:num_atom] = structure.scaled_coords.copy()

    numbers = np.zeros(num_atom * 4, dtype=int)
    numbers[:num_atom] = structure.atomic_numbers.copy()
    num_atom_bravais = spg.refine_cell(lattice,
                                       positions,
                                       numbers,
                                       num_atom,
                                       symprec,
                                       angle_tolerance)

    positions %= 1.0
    positions %= 1.0

    if num_atom_bravais > 0:
        structure.cell = lattice.T
        structure.set_natoms(num_atom_bravais)
        for a, z, c in zip(structure.atoms,
                           numbers[:num_atom_bravais], 
                           positions[:num_atom_bravais]):
            a.element_id = zdict[z]
            a.coord = c
            a.direct = True
        return structure
    else:
        return structure


def find_primitive(structure, symprec=1e-5, angle_tolerance=-1.0):
    """
    A primitive cell in the input cell is searched and returned
    as an object of Atoms class.
    If no primitive cell is found, (None, None, None) is returned.
    """
    lattice = structure.cell.T.copy()
    positions = structure.scaled_coords.copy()
    numbers = structure.atomic_numbers.copy()

    num_atom_prim = spg.primitive(lattice,
                                  positions,
                                  numbers,
                                  symprec,
                                  angle_tolerance)
    positions %= 1.0
    positions %= 1.0
    if num_atom_prim > 0:
        structure.cell = lattice.T
        structure.set_natoms(num_atom_prim)
        for a, z, c in zip(structure.atoms,
                           numbers[:num_atom_prim], 
                           positions[:num_atom_prim]):
            a.element_id = zdict[z]
            a.coord = c
            a.direct = True
        return structure
    else:
        return structure

def equivalent_reflections(hkl, rotations):
    """Return all equivalent reflections to the list of Miller indices
    in hkl.

    Example:
    >>> equivalent_reflections([[0, 0, 2]], rotations)
    array([[ 0,  0, -2],
           [ 0, -2,  0],
           [-2,  0,  0],
           [ 2,  0,  0],
           [ 0,  2,  0],
           [ 0,  0,  2]])
    """
    hkl = np.array(hkl, dtype='int', ndmin=2)
    n, nrot = len(hkl), len(rotations)
    R = rotations.transpose(0, 2, 1).reshape((3*nrot, 3)).T
    refl = np.dot(hkl, R).reshape((n*nrot, 3))
    ind = np.lexsort(refl.T)
    refl = refl[ind]
    diff = np.diff(refl, axis=0)
    mask = np.any(diff, axis=1)
    return np.vstack((refl[mask], refl[-1,:]))

def normalised_reflections(hkl, rotations):
    """Returns an array of same size as *hkl*, containing the
    corresponding symmetry-equivalent reflections of lowest
    indices.

    Example:
    >>> sg.symmetry_normalised_reflections([[2, 0, 0], [0, 2, 0]])
    array([[ 0,  0, -2],
           [ 0,  0, -2]])
    """
    hkl = np.array(hkl, dtype=int, ndmin=2)
    normalised = np.empty(hkl.shape, int)
    R = rotations.transpose(0, 2, 1)
    for i, g in enumerate(hkl):
        gsym = np.dot(R, g)
        j = np.lexsort(gsym.T)[0]
        normalised[i,:] = gsym[j]
    return normalised

def parse_sitesym(sitesym, sep=','):
    rot = np.zeros((3, 3), dtype='int')
    trans = np.zeros(3)
    for i, s in enumerate (sitesym.split(sep)):
        s = s.lower().strip()
        while s:
            sign = 1
            if s[0] in '+-':
                if s[0] == '-':
                    sign = -1
                s = s[1:]
            if s[0] in 'xyz':
                j = ord(s[0]) - ord('x')
                rot[i, j] = sign
                s = s[1:]
            elif s[0].isdigit() or s[0] == '.':
                n = 0
                while n < len(s) and (s[n].isdigit() or s[n] in '/.'):
                    n += 1
                t = s[:n]
                s = s[n:]
                trans[i] = float(frac.Fraction(t))
            else:
                raise ValueError('Failed to parse symmetry of %s' % (sitesym))
    return rot, trans



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from CifFile import ReadCif, CifFile, CifBlock
import warnings
import itertools
import itertools

import qmpy.materials.structure as strx
import qmpy.materials.composition as comp
import qmpy.analysis.symmetry as sym
import qmpy.data.experiment as expt
from qmpy.utils import *

def get_value(number):
    '''Attempt to return a float from a cif number value.'''
    
    if number == '.':
        return 0.0
    try:
        return float(number)
    except ValueError:
        pass
    number = number.split('(')[0]
    return float(number)

def get_atom(cba):
    '''Convert a _atom loop to an Atom'''
    atom = strx.Atom()
    atom.element_id = cba._atom_site_label.strip('0123456789')
    if atom.element_id == 'D':
        atom.element_id = 'H'
    atom.x = get_value(cba._atom_site_fract_x)
    atom.y = get_value(cba._atom_site_fract_y)
    atom.z = get_value(cba._atom_site_fract_z)
    atom.occupancy = get_value(cba._atom_site_occupancy)
    atom.cif_label = cba._atom_site_type_symbol
    return atom

def get_spacegroup(cb):
    '''Create a Spacegroup from a CifBlock'''
    rots, trans = [], []
    for sym in cb.GetLoop('_symmetry_equiv_pos_site_id'):
        r, t = parse_sitesym(sym[1])
        rots.append(r)
        trans.append(t)
    return rots, trans

def get_lattice_parameters(cb):
    '''Return a tuple of lattice parameters from a CifBlock.'''
    return (get_value(cb.get('_cell_length_a')),
            get_value(cb.get('_cell_length_b')),
            get_value(cb.get('_cell_length_c')),
            get_value(cb.get('_cell_angle_alpha')),
            get_value(cb.get('_cell_angle_beta')),
            get_value(cb.get('_cell_angle_gamma')))

def get_reference(cif_block):
    # reference
    reference = expt.Reference()
    journal, new = expt.Journal.objects.get_or_create(
            code=cif_block.get('_citation_journal_id_ASTM')[0])
    if new:
        journal.name = cif_block.get('_citation_journal_full')[0]
    authors = []
    for name in cif_block.get('_publ_author_name'):
        authors.append(expt.Author.from_name(name))
    reference.authors = authors
    reference.journal = journal
    reference.volume = cif_block.get('_citation_journal_volume')[0]
    reference.year = cif_block.get('_citation_year')[0]
    reference.first_page = cif_block.get('_citation_page_first')[0]
    reference.last_page = cif_block.get('_citation_page_last')[0]
    reference.title = cif_block.get('_publ_section_title')
    return reference

def read_cif_block(cb):
    s = strx.Structure()
    s.cell = latparams_to_basis(get_lattice_parameters(cb))
    sym_ops = zip(*get_spacegroup(cb))
    cl = cb.GetLoop('_atom_type_symbol')
    ox_data = set((l._atom_type_symbol , 
        get_value(l._atom_type_oxidation_number)) for l in cl )
    ox_data = dict(ox_data)

    atoms = []
    for atom in cb.GetLoop('_atom_site_type_symbol'):
        na = get_atom(atom)
        na.ox = ox_data[atom._atom_site_type_symbol]
        for rot, trans in sym_ops:
            coord = np.dot(rot, na.coord) + trans
            coord %= 1.0
            a = na.copy()
            a.coord = coord
            if any( a == oa for oa in atoms ):
                continue
            atoms.append(a)
    s.atoms = atoms
    s.composition = comp.Composition.get(s.comp)
    
    ## meta data
    s.r_val = float(cb.get('_refine_ls_R_factor_all', 0.0))
    s.temperature = float(cb.get('_cell_measurement_temperature', 0.0))
    s.pressure = float(cb.get('_cell_measurement_pressure', 0.0))
    if cb.get('_chemical_name_structure_type'):
        s.prototype = strx.Prototype.get(cb.get('_chemical_name_structure_type'))
    s.reported_composition = strx.Composition.get(
                   parse_comp(cb.get('_chemical_formula_sum')))
    s.reference = get_reference(cb)
    s.input_format = 'cif'
    return s

def read_cif(cif_file, grammar='1.1'):
    '''
    Takes a CIF format file, and returns a Structure object. Applies all
    symmetry operations in the CIF to the atoms supplied with the structure. If
    these are not correct, the structure will not be either. If the CIF
    contains more than one file, the return will be a list. If not, the return
    will be a single structure (not in a len-1 list).

    Example:
    >>> s = read_cif('test.cif')
    '''
    warnings.catch_warnings()
    warnings.simplefilter("ignore")
    warnings.warn("deprecated", DeprecationWarning)
    cf = ReadCif(cif_file, grammar=grammar)
    structures = []
    for key in cf.keys():
        structures.append(read_cif_block(cf[key]))
    if len(structures) == 1:
        return structures[0]
    else:
        return structures

def add_comp_loop(structure, cb):
    c_cols = [[ '_atom_type_symbol', '_atom_type_oxidation_number' ]]
    c_data = [[ [ str(s) for s in structure.species ],
                [ str(s.ox_format) for s in structure.species ] ]]
    cb.AddCifItem(( c_cols, c_data ))

def add_cell_loop(structure, cb):
    a, b, c, alpha, beta, gamma = structure.lat_params
    cb['_cell_length_a'] = '%08f' % a
    cb['_cell_length_b'] = '%08f' % b
    cb['_cell_length_c'] = '%08f' % c
    cb['_cell_angle_alpha'] = '%08f' % alpha
    cb['_cell_angle_beta'] = '%08f' % beta
    cb['_cell_angle_gamma'] =  '%08f' % gamma
    cb['_cell_volume'] = '%08f' % structure.volume

def add_symmetry_loop(structure, cb, wrap=False):
    structure.group_atoms_by_symmetry()
    data = sym.get_symmetry_dataset(structure)
    eqd = dict( (i, e) for i, e in enumerate(data['equivalent_atoms']) )
    cb['_symmetry_space_group_name_H-M'] = str(data['international'])
    cb['_symmetry_Int_Tables_number'] = str(data['number'])
    ss_cols = [[ '_symmetry_equiv_pos_site_id', '_symmetry_equiv_pos_as_xyz']]
    ss_data = [[ [ str(i+1) for i in range(len(data['rotations'])) ],
                 [ str(sym.Operation.get((r, t))) for r, t in 
                             zip(data['rotations'], data['translations']) ] ]]
    cb.AddCifItem(( ss_cols, ss_data ))
    a_rows = [[ '_atom_site_label', '_atom_site_type_symbol',
                '_atom_site_fract_x', '_atom_site_fract_y', '_atom_site_fract_z', 
                '_atom_site_Wyckoff_symbol',
                '_atom_site_occupancy' ]]
    a_data = [ [ str('%s%d' % (a.element_id, eqd[i])) 
                      for i, a in enumerate(structure)],
                [ str('%s%+d' % (a.element_id, 
                    a.ox if a.ox else 0 ))
                                 for a in structure ],
                [ '%08f' % a.x for a in structure ], 
                [ '%08f' % a.y for a in structure ], 
                [ '%08f' %a.z for a in structure ], 
                [ str(data['wyckoffs'][i]) for i in range(len(structure)) ],
                [ '%08f' % a.occupancy for a in structure ] ]

    if wrap:
        tmparr = np.array(a_data).T.tolist()
        for a in list(tmparr):
            for i, j, k in itertools.product([0, 1], [0, 1], [0, 1]):
                if i == 0 and j == 0 and k == 0:
                    continue
                if ( float(a[2])+i <= 1 and 
                     float(a[3])+j <= 1 and
                     float(a[4])+k <= 1):
                    tmparr.append([ a[0], a[1], 
                        '%08f' % (float(a[2])+i),
                        '%08f' % (float(a[3])+j),
                        '%08f' % (float(a[4])+k),
                        a[5], a[6]])
        tmparr = sorted(tmparr, key=lambda x: x[0])
        a_data = np.array(tmparr).T.tolist()
    cb.AddCifItem(( a_rows, [a_data] ))

def make_cif_block(structure, wrap=False):
    cb = CifBlock()
    add_cell_loop(structure, cb)
    add_symmetry_loop(structure, cb, wrap=wrap)
    add_comp_loop(structure, cb)
    return cb

def write_cif(structures, wrap=False):
    cif = CifFile()
    if not isinstance(structures, list):
        structures=[structures]
    for i, structure in enumerate(structures):
        cif['structure_%d' % i ] = make_cif_block(structure, wrap=wrap)
    return str(cif)



import json
import os.path
import time
import sys
import logging
from datetime import datetime, timedelta

from qmpy.computing.queue import Task, Job
from qmpy.computing.resources import *
from qmpy.utils.daemon import Daemon
from qmpy.analysis.vasp.calculation import VaspError

logger = logging.getLogger(__name__)

def check_die():
    if os.path.exists('/home/sjk648/die'):
        exit(0)

class JobManager(Daemon):
    def run(self):
        while True:
            jobs = Job.objects.filter(state=1, account__host__state=1)
            for job in jobs:    
                check_die()
                if job.is_done():
                    print job
                    job.collect()
                    job.save()
            for i in range(20):
                time.sleep(1)
                check_die()

class TaskManager(Daemon):
    def run(self, project=None, path=None, id=None):
        while True:
            check_die()

            tasks = Task.objects.none()
            while not tasks.exists():
                if project:
                    tasks = Project.get(project).task_set.filter(state=0)
                else:
                    tasks = Task.objects.filter(project_set__state=1, state=0)
                tasks = tasks.exclude(entry__meta_data__type='hold')
                if tasks.exists():
                    break

                Project.objects.filter(state=0).update(state=1)
                for i in range(120):
                    time.sleep(1)
                    check_die()

            time.sleep(10)
            for task in tasks.order_by('priority')[:50]:
                if not task.eligible_to_run:
                    task.state = -1
                    task.save()
                    continue

                print task

                try:
                    jobs = task.get_jobs()
                except VaspError:
                    print '    -VASP Error.'
                    task.state = -1
                    task.save()
                    continue
                except Exception, err:
                    task.state = -1
                    task.save()
                    print '   -Unknown error!', err
                    continue

                for job in jobs:
                    while not job.state == 1:
                        try:
                            job.submit()
                        except Exception:
                            print '  -waiting 30 seconds to resubmit'
                            time.sleep(30)
                    job.save()
                task.save()
                task.entry.save()





from numpy.linalg import solve, norm
import time
import copy
from collections import defaultdict

from django.db import models

from element import Element, Species
from atom import Atom, Site
from composition import Composition
from qmpy.utils import *
from qmpy.data.meta_data import *
from qmpy.analysis.symmetry import Spacegroup
from qmpy.analysis.symmetry import *
from qmpy.analysis.pdf import get_pdf_uri

class StructureError(Exception):
    '''Structure related problem'''

@add_meta_data('comment')
@add_meta_data('keyword')
class Structure(models.Model, object):
    '''
    Structure model. Principal attributes are a lattice and basis set.

    Database Attributes:
        entry               : Entry which the structure belongs to.
        composition         : Composition of the structure.
        element_set         : django QuerySet of Element objects present in
                              the structure.
        species_set         : django QuerySet of Species objects present in the
                              structure. Species are Elements with a given
                              charge state. 
        reference           : Reference object, containing the author, journal,
                              title, etc...
        label               : Description of the structure
        prototype           :
        natoms              :
        ntypes              :
        nsites              : 
        
        x1, x2, x3
        y1, y2, y3
        z1, z2, z3 - elements of 3x3 lattice vector matrix

        sxx, syy, szz
        sxy, syz, szx - Stress elements

        volume              : total volume, in A^3/UC (float)
        volume_pa           : volume per atom, in A^3/atom (float)

        energy              : total energy, in eV/UC (float)
        energy_pa           : energy/atom, in eV/atom (float)
        delta_e    : formation energy in eV/atom (float)
        meta_stability      : Distance relative to the hull, in the absence of
                              this phase. Negative values indicate stability,
                              positive values, instability. Deeply negative
                              values indicate a strong preference to competing
                              phases, very positive values indicate it is far
                              less stable than competing phases. (float)

        spacegroup          : Spacegroup object

    '''
    # model relationships
    entry = models.ForeignKey('Entry', null=True)
    element_set = models.ManyToManyField('Element')
    species_set = models.ManyToManyField('Species')
    meta_data = models.ManyToManyField('MetaData')
    reference = models.ForeignKey('Reference', blank=True, null=True)
    label = models.CharField(blank=True, max_length=20)
    prototype = models.ForeignKey('Prototype', null=True, blank=True,
                                               related_name='+')
    step = models.IntegerField(blank=True, null=True)
    measured = models.BooleanField(default=False)

    composition = models.ForeignKey('Composition', null=True, 
                                    related_name='structure_set')
    natoms = models.IntegerField(null=True, blank=True)
    nsites = models.IntegerField(null=True, blank=True)
    ntypes = models.IntegerField(null=True, blank=True)

    ## lattice
    x1 = models.FloatField()
    x2 = models.FloatField()
    x3 = models.FloatField()
    y1 = models.FloatField()
    y2 = models.FloatField()
    y3 = models.FloatField()
    z1 = models.FloatField()
    z2 = models.FloatField()
    z3 = models.FloatField()

    volume = models.FloatField(blank=True, null=True)
    volume_pa = models.FloatField(blank=True, null=True)

    sxx = models.FloatField(default=0)
    syy = models.FloatField(default=0)
    szz = models.FloatField(default=0)
    sxy = models.FloatField(default=0)
    syz = models.FloatField(default=0)
    szx = models.FloatField(default=0)

    ## symmetry
    spacegroup = models.ForeignKey('Spacegroup', blank=True,
            null=True)

    ## computed quantities
    energy = models.FloatField(blank=True, null=True)
    energy_pa = models.FloatField(blank=True, null=True)
    magmom = models.FloatField(blank=True, null=True)
    magmom_pa = models.FloatField(blank=True, null=True)
    delta_e = models.FloatField(blank=True, null=True)
    meta_stability = models.FloatField(blank=True, null=True)

    # cache
    _reciprocal_lattice = None
    _distinct_atoms = []
    _magmoms = []

    class Meta:
        app_label = 'qmpy'
        db_table = 'structures'

    # builtins
    def __eq__(self, other):
        if self.comp != other.comp:
            return False
        if not np.allclose(self.cell, other.cell):
            return False
        if not np.allclose(self.scaled_coords, other.scaled_coords):
            return False
        return True

    def __str__(self):
        if not self.composition_id is None:
            return self.composition_id
        else:
            return format_comp(reduce_comp(self.comp))

    def __getitem__(self, i):
        return self.atoms[i]

    def __len__(self):
        return len(self.atoms)

    # django overrides
    def save(self,*args, **kwargs):
        if not self.composition:
            self.composition = Composition.get(self.comp)

        self.symmetrize()
        self.natoms = len(self.atoms)
        self.ntypes = len(self.comp.keys())
        super(Structure, self).save(*args, **kwargs)
        #for a in self.atoms:
        #    a.structure = self
        self.atom_set = self.atoms
        #for s in self.sites:
        #    s.structure = self
        self.site_set = self.sites
        self.element_set = self.elements
        self.species_set = self.species
        self.meta_data = self.comment_objects

    # django accessors/aggregators
    _atoms = None
    @property
    def atoms(self):
        if self._atoms is None:
            if not self.id:
                self._atoms = []
            else:
                self._atoms = self.atom_set.all()
        return self._atoms

    @atoms.setter
    def atoms(self, atoms):
        self._atoms = sorted(atoms)
        self.natoms = len(atoms)
        self.ntypes = len(self.comp)
        for a in atoms:
            a.structure = self

    _sites = None
    @property
    def sites(self):
        if self._sites is None:
            self._sites = self.get_sites()
        return self._sites

    @sites.setter
    def sites(self, sites):
        self._sites = sites
        for s in sites:
            s.structure = self
        self.nsites = len(sites)

    @property
    def elements(self):
        '''List of Elements'''
        return [ Element.get(e) for e in self.comp.keys() ]

    @property
    def species(self):
        '''List of species'''
        return [ Species.get(s) for s in self.spec_comp.keys() ]

    @property
    def stresses(self):
        '''Calculated stresses, a numpy.ndarray of shape (6,)'''
        return np.array([self.sxx, self.syy, self.szz,
            self.sxy, self.syz, self.szx ])
        
    @stresses.setter
    def stresses(self, vector):
        self.sxx, self.syy, self.szz = vector[0:3]
        self.sxy, self.syx, self.szx = vector[3:6]

    # calculated properties
    @property
    def volume(self):
        '''Volume, calculated from the triple product of self.cell'''
        b1, b2, b3 = self.cell
        return abs(np.dot(np.cross(b1, b2), b3))

    @volume.setter
    def volume(self, value):
        scale = value/self.volume
        self.cell = self.cell * scale**(1/3.)

    @property
    def volume_pa(self):
        '''Volume, normalized per atom.'''
        return self.volume/self.natoms

    @property
    def lat_param_dict(self):
        '''Dictionary of lattice parameters.'''
        v1, v2, v3 = self.cell
        a = norm(v1)
        b = norm(v2)
        c = norm(v3)
        alpha = np.arccos(np.dot(v2, v3)/(b*c))*180/np.pi
        beta = np.arccos(np.dot(v1, v3)/(a*c))*180/np.pi
        gamma = np.arccos(np.dot(v1, v2)/(a*b))*180/np.pi
        return { 'a': a, 'b':b, 'c':c,
                'alpha': alpha, 'beta':beta, 'gamma':gamma }

    @property
    def lat_params(self):
        '''Tuple of lattice parameters (a, b, c, alpha, beta, gamma).'''
        v1, v2, v3 = self.cell
        a = norm(v1)
        b = norm(v2)
        c = norm(v3)
        alpha = np.arccos(np.dot(v2, v3)/(b*c))*180/np.pi
        beta = np.arccos(np.dot(v1, v3)/(a*c))*180/np.pi
        gamma = np.arccos(np.dot(v1, v2)/(a*b))*180/np.pi
        return a, b, c, alpha, beta, gamma

    @property
    def cell(self):
        '''Lattice vectors, 3x3 numpy.ndarray.'''
        return np.array([
                [self.x1, self.x2, self.x3],
                [self.y1, self.y2, self.y3],
                [self.z1, self.z2, self.z3]])

    @cell.setter
    def cell(self, cell):
        self.x1, self.x2, self.x3 = cell[0]
        self.y1, self.y2, self.y3 = cell[1]
        self.z1, self.z2, self.z3 = cell[2]
        self._inv = None

    @property
    def atomic_numbers(self):
        '''List of atomic numbers, length equal to number of atoms.'''
        return np.array([ atom.element.z for atom in self.atoms ])

    @property
    def atom_types(self):
        '''List of atomic symbols, length equal to number of atoms.'''
        return np.array([ atom.element_id for atom in self.atoms ])

    @property
    def species_types(self):
        '''List of species, length equal to number of atoms.'''
        return np.array([ atom.species_id for atom in self.atoms ])

    # methods
    def symmetrize(self):
        '''
        Analyze the symmetry of the structure. Uses spglib to find the
        symmetry. 

        Sets:
            self.spacegroup -> Spacegroup
            self.uniq_sites -> list of unique Sites
            
            for each atom:
                atom.wyckoff -> WyckoffSite
            for each site:
                site.multiplicity -> int

        '''
        dataset = get_symmetry_dataset(self)
        self.spacegroup = Spacegroup.objects.get(pk=dataset['number'])
        for i, site in enumerate(self.sites):    
            site.wyckoff = self.spacegroup.get_site(dataset['wyckoffs'][i])
            site.structure = self
        counts = defaultdict(int)
        for e in dataset['equivalent_atoms']:
            counts[e] += 1
        self.uniq_sites = []
        for ind, mult in counts.items():
            site = self.sites[ind]
            site.multiplicity = mult
            self.uniq_sites.append(site)

    def desymmetrize(self):
        '''
        Using self.spacegroup and self.uniq_sites, creates additional sites and
        atoms as indicated by the symmetry of the spacegroup.
        
        '''
        self.atoms = self.spacegroup.equivalent_sites(self)

    def add_atom(self, atom):
        for a in self.atoms:
            if (np.allclose(a.coord, atom.coord) and 
                    a.element_id == atom.element_id ):
                return
        self._atoms.append(atom)
        self.spacegroup = None

    def set_magnetism(self, type, scheme='primitive'):
        '''
        Assigns magnetic moments to all atoms in accordance with the specified
        magnetism scheme.

        Allowed schemes:
            none    - all magnetic moments are set to None
            "ferro" - all magnetic moments with partial d or f shells are
                      assigned an initial magnetic moment of 5 or 7
                      respectively
            "anti"  - raises NotImplementedError

        '''
        if type == 'none':
            for atom in self.atoms:
                atom.magmom = 0
                if atom.id is not None:
                    atom.save()
        if type == 'ferro':
            for atom in self.atoms:
                if atom.element.d_elec > 0 and atom.element.d_elec < 10:
                    atom.magmom = 5
                elif atom.element.f_elec > 0 and atom.element.f_elec < 14:
                    atom.magmom = 7
                else:
                    atom.magmom = 0
                if atom.id is not None:
                    atom.save()
        elif type == 'anti-ferro':
            raise NotImplementedError
        self.spacegroup = None

    @property
    def comp(self):
        '''Composition dictionary.'''
        comp = {}
        for atom in self.atoms:
            elt = atom.element_id
            comp[elt] = comp.get(elt, 0) + atom.occupancy
        return comp

    @property
    def spec_comp(self):
        '''Species composition dictionary.'''
        spec_comp = {}
        for atom in self.atoms:
            spec = atom.species
            spec_comp[spec] = spec_comp.get(spec, 0) + atom.occupancy
        return spec_comp

    @property
    def name(self):
        '''Unformatted name.'''
        return format_comp(self.comp)

    @property
    def unit_comp(self):
        return dict(name_to_comp(self.composition_id, special='unit'))

    @property
    def coords(self):
        '''numpy.ndarray of atom coordinates of shape (3, natoms).'''
        return np.array([ atom.coord for atom in self.atoms ])

    @property
    def magmoms(self):
        '''numpy.ndarray of magnetic moments of shape (natoms,).'''
        return np.array([ atom.magmom for atom in self.atoms ])

    @property
    def site_coords(self):
        '''numpy.ndarray of site coordinates of shape (3, nsites).'''
        return np.vstack([ s.coord for s in self.sites ])

    @property
    def site_labels(self):
        '''
        Labels for all sites. For sites with single occupancy returns the
        atomic symbol, for sites with multiple occupancies returns "(A,B)".

        '''
        return np.array([ str(s) for s in self.sites ])

    @property
    def scaled_coords(self):
        '''Return atomic positions in DIRECT coordinates.'''
        inv = self.inv.T
        coords = []
        for atom in self.atoms:
            if not atom.direct:
                coords.append(np.dot(inv, atom.coord))
            else:
                coords.append(atom.coord)
        coords = np.array(coords)
        coords %= 1.0
        coords %= 1.0
        return coords

    @property
    def direct_coords(self):
        '''Return atomic positions in DIRECT coordinates.'''
        return self.scaled_coords

    @property
    def cartesian_coords(self):
        '''Return atomic positions in cartesian coordinates.'''
        coords = []
        for atom in self.atoms:
            if atom.direct:
                coords.append(np.dot(atom.coord, self.cell))
            else:
                coords.append(atom.coord)
        return np.array(coords)

    @property
    def forces(self):
        '''numpy.ndarray of forces on atoms.'''
        forces = []
        for atom in self.atoms:
            forces.append(atom.forces)
        return np.array(forces)

    @forces.setter
    def forces(self, forces):
        for forces, atom in zip(forces, self.atoms):
            atom.forces = force

    @property
    def reciprocal_lattice(self):
        '''Reciprocal lattice of the structure.'''
        if self._reciprocal_lattice is None:
            self._reciprocal_lattice = np.linalg.inv(self.cell).T
        return self._reciprocal_lattice

    _inv = None
    @property
    def inv(self):
        '''
        Precalculates the inverse of the lattice, for faster conversion
        between cartesian and direct coordinates.
        
        '''
        if self._inv is None:
            self._inv = np.linalg.inv(self.cell)
        return self._inv

    @property
    def relative_rec_lat(self):
        rec_lat = self.reciprocal_lattice
        rec_mags = [ norm(rec_lat[0]), norm(rec_lat[1]), norm(rec_lat[2])]
        r0 = min(rec_mags)
        return np.array([ np.round(r/r0, 4) for r in rec_mags ])

    @property
    def pdf_uri(self):
        return get_pdf_uri(self)

    def get_distance(self, atom1, atom2):
        '''Return the shorted distance between two atoms in the structure.'''
        if isinstance(atom1, int):
            atom1 = self.atoms[atom1]
        if isinstance(atom2, int):
            atom2 = self.atoms[atom2]
        if atom1.direct:
            v1 = np.dot(atom1.coord, self.cell)
        else:
            v1 = atom1.coord
        if atom2.direct:
            v2 = np.dot(atom2.coord, self.cell)
        else:
            v2 = atom2.coord
        vec = v1 - v2
        Dr = solve(self.cell.T, vec)
        D = np.dot(Dr - np.round(Dr), self.cell)
        return norm(D)

    def copy(self):
        '''
        Create a complete copy of the structure, with any primary keys
        removed, so it is not associated with the original.

        '''
        new = copy.deepcopy(self)
        new.id = None
        new.label = ''
        atom_set = []
        for atom in self.atoms:
            new_atom = atom.copy()
            new_atom.id = None
            atom_set.append(new_atom)
        new.atoms = atom_set
        return new

    @property
    def similar(self):
        return Structure.objects.filter(natoms=self.natoms,
                composition=self.composition, 
                label=self.label)

    def set_natoms(self, n):
        '''Set self.atoms to n blank Atoms.'''
        self.atoms = [ Atom() for i in range(n) ]

    def make_conventional(self):
        '''Uses spglib to convert to the conventional cell.'''
        refine_cell(self)

    def make_primitive(self):
        '''Uses spglib to convert to the primitive cell.'''
        find_primitive(self)

    def get_sites(self, tol=1e-3):
        '''
        From self.atoms, creates a list of Sites. Atoms which are closer
        than tol from one another are considered on the same site.

        '''
        if not any([ a.occupancy < 1 for a in self.atoms ]):
            self._sites = [ a.as_site() for a in self.atoms ]
            return self._sites
        d_sites = []
        for atom in self:
            for site in d_sites:
                if atom.is_on(site):
                    site.atoms.append(atom)
                    break
            else:
                site = atom.as_site()
                site.structure = self
                d_sites.append(site)
        self._sites = d_sites
        return self._sites

    def group_atoms_by_symmetry(self):
        '''Sort self.atoms according to the site they occupy.'''
        eq = get_symmetry_dataset(self)['equivalent_atoms']
        resort = np.argsort(eq)
        self._atoms = list(np.array(self._atoms)[resort])

class Prototype(models.Model):
    name = models.CharField(max_length=63, primary_key=True)
    structure = models.ForeignKey(Structure, related_name='+', 
                                  blank=True, null=True)
    composition = models.ForeignKey('Composition', blank=True, null=True)

    class Meta:
        app_label = 'qmpy'
        db_table = 'prototypes'

    @classmethod
    def get(cls, name):
        obj, new = cls.objects.get_or_create(name=name)
        if new:
            obj.save()
        return obj



from django.db import models

from qmpy.utils import *

class Element(models.Model):
    '''
    Database Attributes:
        =-- Identification --=
        z                   : atomic number (int)
        name                : full atomic name (str)
        symbol              : atomic symbol (str)
        
        =-- Periodic table location --=
        group               : group in the periodic table (int)
        period              : period in the periodic table (int)

        =-- Physical characteristics --=
        mass                : Atomic mass, in AMU (float)
        density             : Density at STP, in g/cm^3 (float)
        volume              : Atomic volume at STP, in A^3/atom (float)
        atomic_radii        : in A (float)
        van_der_waals radii : in A (float)
        covalent_radii      : in A (float)

        =-- Thermodynamic characteristics --=
        melt                : melting point in K (float)
        boil                : boiling point in K (float)
        specific_heat       : C_p (float)

        =-- Electronic structure --=
        electronegativity   : Pauling electronegativity (float)
        ion_energy          : Energy to ionize the first electron (float)
        s_elec              : # of s electrons (int)
        p_elec              : # of p electrons (int)
        d_elec              : # of d electrons (int)
        f_elec              : # of f electrons (int)

        =-- Additional information --=
        production          : Annual tons of element produced (float)
        radioactive         : (bool)

    Methods:
        (classmethod)
        get(identity)
            Returns Element corresponding to identity. Identity can be
            elemental name, symbol or atomic number. Can also be a species
            name, i.e. Fe3+, O2-, which which case it only returns the element.

        species_distribution()
            Retuns dict of {oxidation_state: # of occurences}
    '''
    ### Identification
    z = models.IntegerField(db_index=True)
    name = models.CharField(max_length=20)
    symbol = models.CharField(max_length=9, primary_key=True)

    ### Periodic table
    group = models.IntegerField()
    period = models.IntegerField()

    ### Phyical characteristics
    mass = models.FloatField()
    density = models.FloatField()
    volume = models.FloatField()
    atomic_radii = models.IntegerField()
    van_der_waals_radii = models.IntegerField()
    covalent_radii = models.IntegerField()

    ### Thermodynamics
    melt = models.FloatField()
    boil = models.FloatField()
    specific_heat = models.FloatField()

    ### Electonic structure
    electronegativity = models.FloatField()
    first_ionization_energy = models.FloatField()
    s_elec = models.IntegerField()
    p_elec = models.IntegerField()
    d_elec = models.IntegerField()
    f_elec = models.IntegerField()

    ### misc
    production = models.FloatField(default=0)
    radioactive = models.BooleanField(default=False)

    class Meta:
        app_label = 'qmpy'
        db_table = 'elements'

    # builtins
    def __str__(self):
        return self.symbol

    # accessor
    @classmethod
    def get(cls, value):
        '''
        Return an element object. Accepts symbols and atomic numbers, or a list
        of symbols/atomic numbers.
        
        Example:
        >>> Element.get('Fe')
        >>> Element.get(26)
        >>> Element.get(['Fe', 'O', 2])
        '''
        if isinstance(value, cls):
            return value
        elif isinstance(value, list):
            return [ cls.get(v) for v in value ]
        elif isinstance(value, int):
            return cls.objects.get(z=value)
        elif isinstance(value, basestring):
            return cls.objects.get(symbol=value)

    # methods
    def species_distribution(self):
        counts = {}
        for s in self.species_set.all():
            counts[s.ox] = s.structure_set.count()
        return counts

class Species(models.Model):
    '''
    Database Attributes:
        name        : Species name, e.g. Fe3+, O2- (str)
        element     : Element object, <Element: Fe>, <Element: O>
        ox          : Oxidation state - accepts non-integer values. (float)

    Convenience Attributes:

    Methods:
        (classmethod)
        get(identity) -> Species
            return Species object from a descriptive value. Accepts many forms
            of identity. Examples inlcude: Fe3+, Fe3, {'element':'Fe', 'ox':3}.
            Also accepts a list of such descriptors, returns list of Species.
    '''
    name = models.CharField(max_length=8, primary_key=True)
    element = models.ForeignKey(Element, blank=True, null=True)
    ox = models.FloatField(blank=True, null=True)

    class Meta:
        app_label = 'qmpy'
        db_table = 'species'

    # builtins
    def __str__(self):
        return str(self.name)

    # accessor
    @classmethod
    def get(cls, value):
        '''
        Return a Species object. 

        Example:
        >>> Species.get('Fe3+')
        >>> Species.get('Fe3')
        >>> Species.get({'element':'Fe', 'state':3})
        >>> Species.get(('Fe', 3))
        >>> Species.get([ 'Fe3+', 'O2-', 'Li1+'])
        '''
        if isinstance(value, cls):
            return value
        elif isinstance(value, basestring):
            spec, new = cls.objects.get_or_create(name=value)
            if new:
                elt, ox = parse_species(value)
                spec.element_id = elt
                spec.ox = ox
                spec.save()
            return spec
        elif isinstance(value, list):
            return [ cls.get(value) for value in list ]

    @property
    def ox_format(self):
        if self.ox is None:
            return 0
        elif is_integer(self.ox):
            return int(self.ox)
        else:
            return float(round(self.ox, 3))
        



'''Stores details about computational resources which are available to qmpy for
doing calculations.

All resource objects - users, hosts, accounts, allocations, and projects - have
a state variable which can take on the following values, with the following
meanings:
    1   :   active, ready to rock
    0   :   no resources available, i.e. will be reset to active again when
            resources free up
    -1  :   manually set to unavailable, i.e. will NOT be reset to active if
            resources are found to be available
    -2  :   bigger failure, also will not reset to active if found to be
            available.

'''

from datetime import datetime
import random
import subprocess
import os
import os.path
import time
import json

from django.db import models
from django.contrib.auth.models import User

class AllocationError(Exception):
    '''Problem with the allocation'''

class SubmissionError(Exception):
    '''Failed to submit a job'''

class Host(models.Model):
    name = models.CharField(max_length=20, db_index=True)
    ip_address = models.CharField(max_length=120)
    hostname = models.CharField(max_length=120)
    binaries = models.TextField(default=json.dumps({}))
    ppn = models.IntegerField(default=8)
    nodes = models.IntegerField(default=30)
    walltime = models.IntegerField(default=3600*24)
    sub_script = models.CharField(max_length=120)
    sub_text = models.TextField(default='/usr/local/bin/qsub')
    check_queue = models.CharField(max_length=180,
            default='/usr/local/maui/bin/showq')
    checked_time = models.DateTimeField(default=datetime.min)
    running = models.TextField(blank=True)
    utilization = models.IntegerField(default=0)

    state = models.IntegerField(default=1)
    class Meta:
        app_label = 'qmpy'
        db_table = 'hosts'

    def __str__(self):
        return self.name

    @classmethod
    def interactive_create(cls):
        host = {}
        host['name'] = raw_input('Hostname:')
        if cls.objects.filter(name=host['name']).exists():
            print 'Host by that name already exists!'
            exit(-1)
        host['ip_address'] = raw_input('IP Address:')
        if cls.objects.filter(ip_address=host['ip_address']).exists():
            print 'Host at that address already exists!'
            exit(-1)
        host['ppn'] = raw_input('Processors per node:')
        host['nodes'] = raw_input('Max nodes to run on:')
        host['sub_script'] = raw_inputs('Command to submit a script '+
                '(e.g. /usr/local/bin/qsub):')
        host['check_queue'] = raw_input('Command for showq (e.g.'+
                '/usr/local/maui/bin/showq):')
        host['sub_text'] = raw_input('Path to qfile template:')
        h = cls(**host)
        h.save()

    @property
    def accounts(self):
        return list(self.account_set.all())

    @property
    def jobs(self):
        jobs = []
        for acct in self.accounts:
            jobs += list(acct.job_set.filter(state=1))
        return jobs

    @property
    def active(self):
        if self.state < 1:
            return False
        elif self.utilization > self.nodes*self.ppn:
            return False
        else:
            return True

    @property
    def percent_utilization(self):
        return 100. * float(self.utilization) / (self.nodes*self.ppn)

    @property
    def utilization(self):
        util = 0
        for acct in self.account_set.all():
            for job in acct.job_set.filter(state=1):
                util += job.ncpus
        return util

    @property
    def qfile(self):
        return open(self.sub_text).read()

    def get_binary(self, key):
        bins = json.loads(self.binaries)
        return bins.get(key, None)

    def check_host(self):
        ret = subprocess.call("ping -c 1 -w 1 %s" % self.ip_address,
                shell=True,
                stdout=open('/dev/null', 'w'),
                stderr=subprocess.STDOUT)
        if ret == 0:
            return True
        else:
            self.state = -2
            self.save()
            return False

    def check_running(self, verbosity=0):
        self.checked_time = datetime.now()
        if self.state < 0:
            self.running = json.dumps({})
            self.save()
            return
        account = random.choice(self.accounts)
        raw_data = account.execute(self.check_queue, verbosity=verbosity)
        running = {}
        if not raw_data:
            return
        for line in raw_data.split('\n'):
            if 'Active Jobs' in line:
                continue
            line = line.split()
            if len(line) != 9:
                continue
            try:
                running[int(line[0])] = {
                        'user':line[1],
                        'state':line[2],
                        'proc':int(line[3])}
            except:
                pass
        self.running = json.dumps(running)
        self.save()
        
    def get_running(self):
        if self.running is not None:
            return json.loads(self.running)
        else:
            return {}

#===============================================================================#

class Account(models.Model):
    user = models.ForeignKey(User)
    host = models.ForeignKey(Host)
    username = models.CharField(max_length=20)
    run_path = models.CharField(max_length=120)

    state = models.IntegerField(default=1)
    class Meta:
        app_label = 'qmpy'
        db_table = 'accounts'

    def __str__(self):
        return '{user}@{host}'.format(user=self.user.username, 
                host=self.host.name)

    @property
    def active(self):
        if self.state < 1:
            return False
        elif not self.host.active:
            return False
        else:
            return True

    def submit(self, path=None, run_path=None, 
            qfile=None, verbosity=0):
        self.execute('mkdir %s' % run_path, verbosity=verbosity,
                ignore_output=True)
        self.copy(folder=path, file='*',
                destination=run_path,
                verbosity=verbosity)
        cmd = 'cd {path} && {sub} {qfile}'.format(
                path=run_path, 
                sub=self.host.sub_script,
                qfile=qfile)
        stdout = self.execute(cmd, verbosity=verbosity)
        jid = int(stdout.split()[0].split('.')[0])
        if self.host.name == 'quest':
            time.sleep(60)
        return jid

    def execute(self, command='exit 0', verbosity=0, ignore_output=False):
        ssh = 'ssh {user}@{host} "{cmd}"'.format(
                user=self.username,
                host=self.host.ip_address,
                cmd=command)

        if verbosity > 0:
            print ssh

        call = subprocess.Popen(ssh, shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE)
        stdout,stderr = call.communicate()

        if verbosity > 0:
            print 'stdout:', stdout
            print 'stderr:', stderr
        if stderr and not ignore_output:
            print 'WARNING:'
            print stderr
        return stdout

    def copy(self, destination=None, to=None, # where to send the stuff
            fr=None, file=None, folder=None, # what to send
            clear_dest_dir=False, verbosity=0, move=False): # some conditions on sending it

        if destination is None:
            destination = self.run_path
        if to is None:
            to = self
        if fr is None:
            if to == 'local':
                fr = self
            else:
                fr = 'local'

        assert (isinstance(to, Account) or to == 'local')
        assert (isinstance(fr, Account) or fr == 'local')

        send_dir = False
        if file is None and folder is None:
            print 'Must specify something to copy'
        elif file is None:
            send_dir = True
        elif folder is None:
            folder = os.path.dirname(file)
            file = os.path.basename(file)
        
        if clear_dest_dir:
            if to == 'local':
                command = subprocess.Popen('rm -f %s/*' % destination,
                                                  stderr=subprocess.PIPE,
                                                  stdout=subprocess.PIPE)
                stdout, stderr = command.communicate()
            else:
                stdout, stderr = self.execute('rm -f %/*' % destination)
            if verbosity:
                print 'output:', stdout
            
        if fr == 'local':
            scp = 'scp '
        else:
            scp = 'scp {user}@{host}:'.format(
                    user=fr.username, host=fr.host.ip_address)

        if not file:
            scp += '-r '

        if send_dir:
            scp += os.path.abspath(folder)
        else:
            scp += '{path}/{file}'.format(
                    path=os.path.abspath(folder), file=file)

        if to == 'local':
            scp += ' '+destination
        else:
            scp += ' {user}@{host}:{path}'.format(
                user=to.username, host=to.host.ip_address, 
                path=os.path.abspath(destination))

        if verbosity > 0:
            print 'copy command:', scp
        cmd = subprocess.Popen(scp,
                shell=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE)
        stdout, stderr = cmd.communicate()
        if verbosity:
            print 'stdout:', stdout
            print 'stderr:', stderr

        if move:
            if send_dir:
                rmcmd = 'rm -rf {path}'.format(path=os.path.abspath(folder))
            else:
                rmcmd = 'rm -f {path}/{file}'.format(file=file,
                        path=os.path.abspath(folder))
            if verbosity > 0:
                print 'wiping source: ', rmcmd
            stdout = fr.execute(rmcmd)
            if verbosity > 0:
                print 'output:', stdout

#===============================================================================#

class Allocation(models.Model):
    name = models.CharField(max_length=50, db_index=True)
    key = models.CharField(max_length=100, default='', blank=True, null=True)

    host = models.ForeignKey(Host, null=True)
    users = models.ManyToManyField(User, null=True)
    state = models.IntegerField(default=1)
    
    class Meta:
        app_label = 'qmpy'
        db_table = 'allocations'

    def __str__(self):
        return self.name
    
    @classmethod
    def create_interactive(cls):
        name = raw_input('Name your allocation:')
        if cls.objects.filter(name=name).exists():
            print 'Allocation by that name already exists!'
            exit(-1)
        host = raw_input('Which cluster is this allocation on?')
        if not Host.objects.filter(name=host).exists():
            print "This host doesn't exist!"
            exit(-1)
        host = Host.objects.get(name=host)
        alloc = cls(name=name, host=host)
        alloc.save()
        print 'Now we will assign users to this allocation'
        for acct in Account.objects.filter(host=host):
            inc = raw_input('Can %s use this allocation? y/n [y]:' % 
                    acct.user.name )
            if inc == 'y' or inc == '':
                alloc.users.add(acct.user)
        print 'If this allocation requires a special password, enter',
        key = raw_input('it now:')
        alloc.key=key
        alloc.save()

    @property
    def active(self):
        if self.state < 1:
            return False
        elif not self.host.active:
            return False
        else:
            return True

    def get_user(self):
        return random.choice(self.users.filter(state=1))

    def get_account(self, users=None):
        if users is None:
            users = self.users.filter(state=1).all()
        user = random.choice(list(users))
        return user.account_set.get(host=self.host)

    @property
    def percent_utilization(self):
        return self.host.percent_utilization

#===============================================================================#

class Project(models.Model):
    name = models.CharField(max_length=60)
    priority = models.IntegerField(default=0)

    users = models.ManyToManyField(User)
    allocations = models.ManyToManyField(Allocation)
    state = models.IntegerField(default=1)
    class Meta:
        app_label = 'qmpy'
        db_table = 'projects'

    def __str__(self):
        return self.name

    @classmethod
    def get(cls, name):
        if isinstance(name, cls):
            return name
        proj, new = cls.objects.get_or_create(name=name)
        if new:
            proj.save()
        return proj

    @property
    def completed(self):
        return self.task_set.filter(state=2)

    @property
    def running(self):
        from queue import Job
        return Job.objects.filter(state=1, task__project_set=self)

    @classmethod
    def interactive_create(cls):
        name = raw_input('Name your project: ')
        if cls.objects.filter(name=name).exists():
            print 'Project by that name already exists!'
            exit(-1)
        proj = cls(name=name)
        proj.save()
        proj.priority = raw_input('Project priority (1-100): ')
        users = raw_input('List project users (e.g. sjk648 jsaal531 bwm291): ')
        for u in users.split():
            if not User.objects.filter(name=u).exists():
                print 'User named', u, 'doesn\'t exist!'
            else:
                proj.users.add(User.objects.get(name=u))

        alloc = raw_input('List project allocations (e.g. byrd victoria b1004): ')
        for a in alloc.split():
            if not Allocation.objects.filter(name=a).exists():
                print 'Allocation named', a, 'doesn\'t exist!'
            else:
                proj.allocations.add(Allocation.objects.get(name=a))

    @property
    def active(self):
        if self.state < 0:
            return False
        elif not any( a.active for a in self.allocations.all() ):
            self.state = 0
            self.save()
            return False
        else:
            if self.state != 1:
                self.state = 1
                self.save()
            return True

    def get_allocation(self):
        available = [ a for a in self.allocations.all() if a.active ]
        if available:
            return random.choice(available)
        else:
            return []



import ConfigParser
import json
from qmpy import INSTALL_PATH
from qmpy.computing.resources import Host, Account, Project, Allocation
from django.contrib.auth.models import User
from qmpy.configuration.resources import *

config = ConfigParser.ConfigParser()
config.read(INSTALL_PATH+'/configuration/site.cfg')

VASP_POTENTIALS = config.get('VASP', 'potential_path')

def sync_resources():
    for host, data in hosts.items():
        h = Host.objects.get_or_create(name=host)[0]
        h.__dict__.update({'check_queue':data['check_queue'],
            'ip_address':data['ip_address'],
            'binaries':json.dumps(data['binaries']),
            'ppn':data['ppn'],
            'nodes':data['nodes'],
            'walltime':data['walltime'],
            'sub_script':data['sub_script'],
            'sub_text':data['sub_text']})

        h.save()

    for user, data in users.items():
        u = User.objects.get_or_create(username=user)[0]
        u.save()
        for host, adata in data.items():
            host = Host.objects.get_or_create(name=host)[0]
            host.save()
            acc = Account.objects.get_or_create(user=u, host=host)[0]
            acc.__dict__.update(**adata)
            acc.save()

    for allocation, data in allocations.items():
        host = Host.objects.get_or_create(name=data['host'])[0]
        host.save()
        alloc = Allocation.objects.get_or_create(name=allocation)[0]
        alloc.host = host
        alloc.key = data.get('key', '')
        alloc.save()
        for user in data['users']:
            user = User.objects.get_or_create(username=user)[0]
            user.save()
            alloc.users.add(user)

    for project, data in projects.items():
        proj = Project.objects.get_or_create(name=project)[0]
        proj.save()

        for user in data['users']:
            user = User.objects.get_or_create(username=user)[0]
            user.save()
            proj.users.add(user)

        for allocation in data['allocations']:
            alloc = Allocation.objects.get_or_create(name=allocation)[0]
            alloc.save()
            proj.allocations.add(alloc)



#!/usr/bin/env python

'''
Task:
    state = -2; error
    state = -1; held
    state = 0; ready to advance
    state = 1; running
    state = 2; done

Job:
    state = -1; error
    state = 0; ready to submit
    state = 1; running
    state = 2; done
'''

from django.db import models
import json
import os.path
import time
from datetime import datetime, timedelta
import random

from resources import Project, Account, Allocation
import qmpy

class TaskError(Exception):
    '''A project was needed but not provided'''

class Task(models.Model):
    module = models.CharField(max_length=60)
    args = models.CharField(max_length=100, blank=True)
    state = models.IntegerField(default=0)
    priority = models.IntegerField(default=50)
    created = models.DateTimeField(blank=True, auto_now_add=True)
    finished = models.DateTimeField(blank=True, null=True)

    entry = models.ForeignKey('Entry')
    project_set = models.ManyToManyField(Project, null=True)

    _projects = None

    class Meta:
        app_label = 'qmpy'
        db_table = 'tasks'

    def save(self, *args, **kwargs):
        super(Task, self).save(*args, **kwargs)
        self.project_set = [ Project.get(p) for p in self.projects ]

    @property
    def projects(self):
        if self._projects is None:
            self._projects = list(self.project_set.all())
        return self._projects

    @projects.setter
    def projects(self, projects):
        self._projects = projects

    @property
    def eligible_to_run(self):
        if self.state != 0:
            return False
        if self.entry.holds:
            return False
        if not any( p.active for p in self.project_set.all()):
            return False
        return True

    def get_project(self):
        projects = [ p for p in self.projects if p.active ]
        return random.choice(projects)

    @staticmethod
    def create(entry, module='standard', args=[],
            priority=None, projects=None): 
        if projects is None:
            projects = entry.projects
        elif isinstance(projects, basestring):
            projects = Project.objects.get(name=projects)
        if priority is None:
            priority = len(entry.input)
        task, created = Task.objects.get_or_create(entry=entry, 
                args=json.dumps(args), module=module)
        if created:
            task.projects = projects
        else:
            task.projects += projects

        task.priority = priority
        return task

    def complete(self):
        self.state = 2
        self.finished = datetime.now()

    def __str__(self):
        return '%s (%s: %s)' % (self.module, self.entry, self.entry.path)

    @property
    def jobs(self):
        return self.job_set.all()

    @property
    def errors(self):
        return self.entry.errors

    def get_jobs(self, project=None, allocation=None, account=None):
        if project is None:
            project = self.get_project()
            while not project:
                time.sleep(30)
                project = self.get_project()
        if allocation is None:
            allocation = project.get_allocation()
            while not allocation:
                time.sleep(30)
                allocation = project.get_allocation()
        if account is None:
            account = allocation.get_account(users=list(project.users.all()))

        calc = self.entry.do(self.module, args=json.loads(self.args))
        jobs = []
        #for calc in calcs:
        if calc.instructions:
            self.state = 1
            new_job = Job.create(
                task=self, 
                allocation=allocation, 
                account=account,
                entry=self.entry,
                **calc.instructions)
            jobs.append(new_job)
            calc.save()
        elif calc.converged:
            self.complete()
        else:
            self.state = -1
        return jobs

class Job(models.Model):
    qid = models.IntegerField(default=0)
    walltime = models.DateTimeField(blank=True)
    path = models.CharField(max_length=200)
    run_path = models.CharField(max_length=200)
    ncpus = models.IntegerField(blank=True)
    created = models.DateTimeField(blank=True, auto_now_add=True)
    finished = models.DateTimeField(blank=True, null=True)
    state = models.IntegerField(default=0)

    task = models.ForeignKey(Task)
    entry = models.ForeignKey('Entry')
    account = models.ForeignKey(Account)
    allocation = models.ForeignKey(Allocation)

    class Meta:
        app_label = 'qmpy'
        db_table = 'jobs'

    @staticmethod
    def create(task=None, allocation=None, entry=None,
            account=None,
            path=None, 
            walltime=3600, serial=None,
            header=None,
            mpi=None, binary=None, pipes=None,
            footer=None):

        if entry is None:
            entry = task.entry

        assert isinstance(allocation, Allocation)
        assert isinstance(task, Task)
        assert path is not None

        if account is None:
            account = allocation.get_account()
        
        job = Job(path=path, walltime=walltime, 
                allocation=allocation,
                account=account,
                entry=entry,
                task=task)

        #if walltime < 3600:
        #    nodes = 1
        #    ppn = int(walltime/3600.*job.account.host.ppn)
        #    walltime = walltime/ppn
        #else:
        #    ppn = job.account.host.ppn
        #    nodes = 1+int(walltime/float(job.account.host.walltime))
        #    walltime = walltime/float(ppn*nodes)
        
        if serial:
            ppn = 1
            nodes = 1
            walltime = 3600*24
        else:
            nodes = 1
            ppn = job.account.host.ppn
            walltime = job.account.host.walltime
            
        binary = job.account.host.get_binary(binary)
        if not binary:
            raise AllocationError

        sec = timedelta(seconds=walltime)
        d = datetime(1,1,1) + sec
        job.walltime = d
        walltime = '%02d:%02d:%02d:%02d' % (
                d.day-1, 
                d.hour, 
                d.minute,
                d.second)

        qp = qmpy.INSTALL_PATH + '/configuration/qfiles/'
        text = open(qp+job.account.host.sub_text+'.q', 'r').read()
        qfile = text.format(
                host=allocation.host.name,
                key=allocation.key, name=job.description,
                walltime=walltime, nodes=nodes, ppn=ppn,
                header=header,
                mpi=mpi, binary=binary, pipes=pipes,
                footer=footer)

        qf=open(job.path+'/auto.q', 'w')
        qf.write(qfile)
        qf.close()
        job.ncpus = ppn*nodes
        job.run_path = job.account.run_path+'/'+job.description
        return job

    @property
    def walltime_expired(self):
        from datetime import datetime, timedelta
        elapsed = datetime.now() - self.created
        if elapsed.total_seconds() > self.walltime:
            return True
        else:
            return False

    @property
    def subdir(self):
        return self.path.replace(self.entry.path, '')

    @property
    def description(self):
        return '{entry}_{subdir}_{uniq}'.format(
                entry=self.entry.id,
                subdir=self.subdir.strip('/').replace('/','_'),
                uniq='-'.join(json.loads(self.task.args)))

    def __str__(self):
        return '%s on %s' % (self.description, self.account)

    def is_done(self):
        if datetime.now() + timedelta(seconds=-120) < self.created:
            return False

        if self.account.host.checked_time < datetime.now()+timedelta(seconds=-120):
            self.account.host.check_running()

        if unicode(self.qid) in self.account.host.get_running():
            return False
        else:
            return True

    def submit(self, verbosity=0):
        self.created = datetime.now()
        self.qid = self.account.submit(path=self.path,
                run_path=self.run_path,
                qfile='auto.q', verbosity=verbosity)
        self.task.state = 1
        self.state = 1

    def collect(self, verbosity=0):
        if self.account.host.state == -1:
            return
        self.task.state = 0
        self.task.save()
        self.state = 2
        self.account.copy(move=True,
                to='local', destination=self.path,
                folder=self.run_path, file='*')
        self.account.execute('rm -rf %s' % self.run_path, ignore_output=True)
        self.finished = datetime.now()




'''qmpy.utils.math

Some mathematical functions which are used periodically.

'''
import numpy as np
from numpy.linalg import norm

def _gcd(a,b):
    '''
    Returns greatest common denominator of two numbers.
    
    Example:
    >>> _gcd(4, 6)
    2
    '''
    while b:
        a,b=b,a%b
    return a

def gcd(numbers):
    '''
    Returns greatest common denominator of an iterator of numbers.

    Example:
    >>> gcd(12, 20, 32)
    4
    '''
    numbers = list(numbers)
    if len(numbers) == 1:
        return numbers[0]
    a=numbers.pop()
    b=numbers.pop()
    tmp_gcd = _gcd(a,b)
    while numbers:
        tmp_gcd = _gcd(tmp_gcd,numbers.pop())
    return tmp_gcd

def is_integer(number, tol=1e-5):
    number = float(number)
    return abs(round(number) - number) < tol
        

def lcm(a,b):
    '''
    Returns least common multiple of two numbers.

    Example:
    >>> lcm(20,10)
    20
    '''
    return a*b/gcd(a,b)

def ffloat(string):
    '''
    In case of fortran digits overflowing and returing ********* this
    fuction will replace such values with 1e9
    '''

    try:
        new_float = float(string)
    except ValueError:
        if '*******' in string:
            new_float = 1e9
        else:
            print string
            return None
    return new_float

def angle(x, y):
    '''Return the angle between two lattice vectors

    Example:

    >>> angle([1,0,0], [0,1,0])
    90.0
    '''
    return np.arccos(np.dot(x,y)/(norm(x)*norm(y)))*180./np.pi

def basis_to_latparams(basis):
    """Returns the lattice parameters [a, b, c, alpha, beta, gamma].
    
    Example:
        
    >>> basis_to_latparams([[3,0,0],[0,3,0],[0,0,5]])
    [3, 3, 5, 90, 90, 90]
    """

    va, vb, vc = basis
    a = np.linalg.norm(va)
    b = np.linalg.norm(vb)
    c = np.linalg.norm(vc)
    alpha = angle(vb, vc)
    beta = angle(vc, va)
    gamma = angle(va, vb)
    return [a, b, c, alpha, beta, gamma]


def latparams_to_basis(latparam, units='degrees'):
    '''Convert a 3x3 basis matrix from the lattice parameters.

    Example:
    >>> latparams_to_basis([3, 3, 5, 90, 90, 90])
    [[3,0,0],[0,3,0],[0,0,5]]

    '''
    basis = []
    a, b, c = latparam[0:3]
    alpha, beta, gamma = latparam[3:6]

    if units == 'degrees':
        # need to convert to radians
        alpha *= np.pi/180
        beta *= np.pi/180
        gamma *= np.pi/180

    basis.append([a, 0, 0])
    basis.append([b*np.cos(gamma), b*np.sin(gamma), 0 ])
    cx = np.cos(beta)
    cy = (np.cos(alpha) - np.cos(beta)*np.cos(gamma))/np.sin(gamma)
    cz = np.sqrt(1. - cx*cx - cy*cy)
    basis.append([c*cx, c*cy, c*cz])
    basis = np.array(basis)
    basis[abs(basis) < 1e-10] = 0
    return basis

def coord_to_bin(coord):
    '''Convert a binary composition to an x-coordinate value:
    returns ( A )
    
    '''
    return coords[0]

def coord_to_gtri(coord):
    '''Convert a ternary composition to an x,y-coordinate pair:
    ( A+B/2, B*3^(1/2)/2 )
    
    '''
    return (coord[0] + coord[1]/2., 
            coord[1]*np.sqrt(3)/2)

def coord_to_gtet(coord):
    '''Convert a quaternary composition to an x,y,z triplet:
    ( A/2+B+C/2, A*3^(1/2)/2 + C*3^(1/2)/6, C*(2/3)^(1/2) )
    
    '''
    return (coord[0]/2 + coord[1] + coord[2]/2,
        coord[0]*np.sqrt(3)/2 + coord[2]*np.sqrt(3)/6.,
        coord[2]*np.sqrt(2./3))

def triple_prod(matrix):
    return np.dot(matrix[0], np.cross(matrix[1], matrix[2]))

def intervals(*args):
    for i in range(len(args)-1):
        print args[i+1] - args[i],



#/usr/bin/env python
# -*- coding: utf-8 -*-

from cif import read_cif, write_cif
from poscar import read_poscar, write_poscar
from ase_mapper import read_ase

class StructureFormatError(Exception):
    '''Problem reading an input file'''

class FormatNotSupportedError(Exception):
    '''The provided input format is not supported'''

def read(source_file, *args, **kwargs):
    if 'cif' in source_file:
        return read_cif(source_file, *args, **kwargs)
    elif ( 'POSCAR' in source_file or
            'CONTCAR' in source_file ):
        return read_poscar(source_file, *args, **kwargs)
    else:
        return read_ase(source_file, *args, **kwargs)
    raise FormatNotSupportedError('The file %s is in an unrecognized format\
            and cannot be read' % source_file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from collections import defaultdict

from django.db import models

from qmpy.materials.element import Element
from qmpy.data.meta_data import *
from qmpy.utils import *

class Composition(models.Model):
    '''
    Database Attributes:

    Convenience Attributes:

    Methods:
    '''
    formula = models.CharField(primary_key=True, max_length=255, db_index=True)
    meta_data = models.ForeignKey('MetaData', null=True)

    element_set = models.ManyToManyField('Element', null=True)
    ntypes = models.IntegerField(null=True)

    ### other stuff
    mass = models.FloatField(blank=True, null=True)

    ### thermodyanamic stuff
    meidema = models.FloatField(blank=True, null=True)
    structure = models.ForeignKey('Structure', blank=True,
            null=True,
            related_name='+')

    _unique = None
    _duplicates = None

    class Meta:
        app_label = 'qmpy'
        db_table = 'compositions'

    # builtins
    def __str__(self):
        return self.formula

    def __eq__(self, other):
        if self.space != other.space:
            return False
        for k in self.space:
            if abs(self.unit_comp[k] - other.unit_comp[k]) > 1e-4:
                return False
        return True

    # django overrides
    def save(self, *args, **kwargs):
        super(Composition, self).save(*args, **kwargs)
        self.element_set = [ Element.get(e) for e in self.comp.keys() ]

    # accessors
    @classmethod
    def get(cls, composition):
        if isinstance(composition, basestring):
            composition = parse_comp(composition)
        comp = reduce_comp(composition)
        comp, created = cls.objects.get_or_create(formula=format_comp(comp))
        if created:
            comp.ntypes = len(comp.comp)
            comp.elements = comp.comp.keys()
        return comp

    @classmethod
    def get_list(cls, bounds):
        space = set()
        if isinstance(bounds, basestring):
            bounds = bounds.split('-')
        if len(bounds) == 1:
            return [Composition.get(bounds[0])]
        for b in bounds:
            bound = parse_comp(b)
            space |= set(bound.keys())
        elts = Element.objects.exclude(symbol__in=space)
        comps = Composition.objects.exclude(element_set__in=elts)
        comps = comps.exclude(entry=None)
        return comps

    # django caching
    _elements = None

    @property
    def elements(self):
        if self._elements is None:
            self._elements = [ Element.get(e) for e in self.comp.keys() ]
        else:
            self._elements = list(self.element_set.all())
        return self._elements

    @elements.setter
    def elements(self, elements):
        self._elements = [ Element.get(e) for e in elements ]

    # calculated properties
    @property
    def distinct(self):
        if self._unique is None:
            self.get_distinct_entries()
        return self._unique

    @property
    def delta_e(self):
        calcs = self.calculation_set.exclude(delta_e=None)
        if not calcs.exists():
            return
        return min(calcs.values_list('delta_e', flat=True))

    @property
    def ndistinct(self):
        return len(self.distinct)

    @property
    def comp(self):
        return parse_comp(self.formula)

    @property
    def unit_comp(self):
        return normalize_comp(self.comp)

    @property
    def red_comp(self):
        return reduce_comp(self.comp)

    @property
    def name(self):
        return format_comp(reduce_comp(self.comp))

    @property
    def latex(self):
        return latex_comp(reduce_comp(self.comp))

    @property
    def html(self):
        return html_comp(reduce_comp(self.comp))

    @property
    def space(self):
        return set(self.comp.keys())

    # methods
    def get_distinct_entries(self):
        unique = {}
        duplicates = defaultdict(list)
        entries = self.entry_set.all()
        for entry in entries:
            sg = entry.input.spacegroup.number
            natoms = entry.input.natoms
            duplicates[(sg, natoms)].append(entry)
            if not (sg, natoms) in unique:
                unique[(sg, natoms)] = entry
            elif entry.delta_e < unique[(sg, natoms)].delta_e:
                unique[(sg, natoms)] = entry
        self._unique = unique
        self._duplicates = duplicates
